import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def linspacestep(start, stop, step=1):
    """
    Returns a :class:`numpy.ndarray` starting as *start*, with elements every *step*, up to at maximum *stop*.
    """
    # Find an integer number of steps
    numsteps = _np.int((stop-start)/step)

    # Do a linspace over the new range
    # that has the correct endpoint
    return _np.linspace(start, start+step*numsteps, numsteps+1)



# Author: Joel Frederico
"""
Calculates various useful quantities in accelerator physics.
"""
__all__ = [
    'BDES2K',
    'K2BDES',
    'fitimageslice'
    ]
from .BDES2K import *  # noqa
from .fitimageslice import fitimageslice



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.pyplot as plt
    import numpy as np
    import gtk.gdk as gdk


def tile():
    """Tiles open figures."""

    figs = plt.get_fignums()

    # Keep track of x, y, size for figures
    x       = 0
    y       = 0
    # maxy    = 0
    toppad  = 21

    size = np.array([0, 0])

    if ( len(figs) != 0 ):
        fig     = plt.figure(figs[0])
        screen  = fig.canvas.window.get_screen()
        screenx = screen.get_monitor_geometry(screen.get_primary_monitor())
        screenx = screenx[2]
    
        fig = plt.figure(figs[0])
        fig.canvas.manager.window.move(x, y)
        maxy = np.array(fig.canvas.manager.window.get_position())[1]
        size = np.array(fig.canvas.manager.window.get_size())
        y    = maxy
        x += size[0]+1
    
        for fig in figs[1:]:
            fig  = plt.figure(fig)
            size = np.array(fig.canvas.manager.window.get_size())
            if ( x+size[0] > screenx ):
                x    = 0
                y    = maxy
                maxy = y+size[1]+toppad
            else:
                maxy = max(maxy, y+size[1]+toppad)
            fig.canvas.manager.window.move(x, y)
            x += size[0] + 1
    
if __name__ == "__main__":
    tile()



import os
import shutil
import shlex
import subprocess


def latexfig(textstr, filename, environment='align*', env_curly=None):
    # ======================================
    # Validate Extension
    # ======================================
    split = os.path.splitext(filename)
    if split[1] != '.pdf':
        raise IOError('Final filename must have extension ''pdf'', requested: {}'.format(filename))

    # ======================================
    # Get final destination
    # ======================================
    final_path = os.path.abspath(filename)
    try:
        os.mkdir(os.path.dirname(final_path))
    except OSError as e:
        if e.errno != 17:
            raise
    
    # ======================================
    # Get current directory
    # ======================================
    cwd = os.getcwdu()

    # ======================================
    # Create info to write
    # ======================================
    template = r'''
\documentclass[10pt]{{article}}
\usepackage{{amssymb, amsmath, booktabs, multirow}}

\pagestyle{{empty}}

\begin{{document}}
\begin{{{environment}}}{env_curly}
{textstr}
\end{{{environment}}}
\end{{document}}
'''
    if env_curly is not None:
        env_curly = '{{{}}}'.format(env_curly)
    else:
        env_curly = ''

    fullwrite = template.format(textstr=textstr, environment=environment, env_curly=env_curly)

    # ======================================
    # Get file names for intermediate files
    # ======================================
    tempdir  = 'temp_latex'
    tempfile = 'temp.tex'
    split = os.path.splitext(tempfile)
    pdffile  = '{}.{}'.format(split[0], 'pdf')
    cropfile = '{}-crop.{}'.format(split[0], 'pdf')

    # ======================================
    # Delete and remake temp directory
    # ======================================
    try:
        shutil.rmtree(tempdir)
    except OSError as e:
        if e.errno is not 2:
            raise
    os.mkdir(tempdir)

    # ======================================
    # Create temporary file
    # ======================================
    f = open(os.path.join(tempdir, tempfile), 'w+')
    f.write(fullwrite)
    f.close()

    # ======================================
    # Compile figure
    # ======================================
    os.chdir(tempdir)

    command = 'latexmk {} -pdf'.format(tempfile)
    args = shlex.split(command)
    subprocess.call(args)

    command = 'pdfcrop {} {}'.format(pdffile, cropfile)
    args = shlex.split(command)
    subprocess.call(args)

    # ======================================
    # Move figure to correct location
    # ======================================
    shutil.move(cropfile, final_path)

    # ======================================
    # Clean up
    # ======================================
    os.chdir(cwd)
    shutil.rmtree(tempdir)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


def chisquare(observe, expect, error, ddof, verbose=True):
    """
    Finds the reduced chi square difference of *observe* and *expect* with a given *error* and *ddof* degrees of freedom.

    *verbose* flag determines if the reduced chi square is printed to the terminal. 
    """
    chisq = 0
    error   = error.flatten()
    observe = observe.flatten()
    expect  = expect.flatten()
    for i, el in enumerate(observe):
        chisq = chisq + _np.power((el - expect[i]) / error[i], 2)

    red_chisq = chisq / (len(observe) - ddof)
    if verbose:
        # print 'Chi-Squared is {}.'.format(chisq)
        print('Reduced Chi-Squared is {}.'.format(red_chisq))

    return red_chisq



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def gaussian(x, mu, sigma):
    """
    .. versionadded:: 1.5

    Gaussian function of the form :math:`\\frac{1}{\\sqrt{2 \\pi}\\sigma} e^{-\\frac{(x-\\mu)^2}{2\\sigma^2}}`.
    """
    return _np.exp(-(x-mu)**2/(2*sigma**2)) / (_np.sqrt(2*_np.pi) * sigma)



import argparse as _argparse
import shlex as _shlex
import subprocess as _subprocess


def pdf2png(file_in, file_out):
    """
    Uses `ImageMagick <http://www.imagemagick.org/>`_ to convert an input *file_in* pdf to a *file_out* png. (Untested with other formats.
    """
    command = 'convert -display 37.5 {} -resize 600 -append {}'.format(file_in, file_out)
    _subprocess.call(_shlex.split(command))

if __name__ == '__main__':
    parser = _argparse.ArgumentParser(description=
            'Creates a tunnel primarily for Git.')
    parser.add_argument('-V', action='version', version='%(prog)s v0.1')
    parser.add_argument('-v', '--verbose', action='store_true',
            help='Verbose mode.')
    parser.add_argument('-i', '--input',
            help='Input file.')
    parser.add_argument('-o', '--output',
            help='Input file.')

    arg = parser.parse_args()

    pdf2png(file_in=arg.input, file_out=arg.output)



"""
The :mod:`scipy <scisalt.scipy>` module contains a few convenience functions mostly designed to make fitting easier.
"""
__all__ = ['chisquare', 'curve_fit_unscaled', 'fft', 'gaussfit']
from .LinLsqFit import *                 # noqa
from .chisquare import *                 # noqa
from .curve_fit_unscaled import *        # noqa
from .fft import *                       # noqa
from .fill_missing_timestamps import *   # noqa
from .gaussfit import *                  # noqa

import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import pyximport
    import numpy as _np
    pyximport.install(setup_args={"include_dirs": _np.get_include()})

    from .hough_ellipse import hough_ellipse



def axesfontsize(ax, fontsize):
    """
    Change the font size for the title, x and y labels, and x and y tick labels for axis *ax* to *fontsize*.
    """
    items = ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels())
    for item in items:
        item.set_fontsize(fontsize)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np

from .chisquare import chisquare
import copy as _copy


class LinLsqFit(object):
    """
    Gets the linear least squares for :math:`\\beta` of a problem given :math:`X_{ij} \\beta_{i} = y_j`.

    As input, it takes *y_unweighted* as the measured :math:`y`, *X_unweighted* for :math:`X`, and *y_error* as the measurement error on :math:`y`.
    """
    _resetlist = ['_X', '_y', '_beta', '_covar', '_chisq_red', '_y_fit']

    def __init__(self, y_unweighted, X_unweighted, y_error=None):

        self._force_recalc()

        self.y_unweighted = y_unweighted
        self.y_error = y_error
        self.X_unweighted = X_unweighted

    # ======================================
    # Resets stored values for calculated
    # quantities
    # ======================================
    def _force_recalc(self):
        # self._X = None
        # self._y = None
        # self._beta = None
        # self._covar = None
        # self._chisq_red = None
        # self._emit = None
        # self._twiss=None
        for resetstr in self._resetlist:
            # print resetstr
            setattr(self, resetstr, None)

    # ======================================
    # y_unweighted
    # ======================================
    @property
    def y_unweighted(self):
        """
        The :math:`y` of the problem :math:`X_{ij} \\beta_{i} = y_j`. Setting this attribute forces a recalculation.
        """
        return self._y_unweighted

    @y_unweighted.setter
    def y_unweighted(self, val):
        self._force_recalc()
        self._y_unweighted = val

    # ======================================
    # y_error
    # ======================================
    @property
    def y_error(self):
        """
        The measured error of :math:`y` of the problem :math:`X_{ij} \\beta_{i} = y_j`. Setting this attribute forces a recalculation.
        """
        return self._y_error

    @y_error.setter
    def _set_y_error(self, val):
        self._force_recalc()
        self._y_error = val
    
    # ======================================
    # X_unweighted
    # ======================================
    @property
    def X_unweighted(self):
        """
        The :math:`X`. Setting this attribute forces a recalculation.
        """
        return self._X_unweighted

    @X_unweighted.setter
    def _set_X_unweighted(self, val):
        self._force_recalc()
        self._X_unweighted = val

    # ======================================
    # X (calculated)
    # ======================================
    @property
    def X(self):
        """
        The :math:`X` weighted properly by the errors from *y_error*
        """
        if self._X is None:
            X = _copy.deepcopy(self.X_unweighted)
            # print 'X shape is {}'.format(X.shape)
            for i, el in enumerate(X):
                X[i, :] = el/self.y_error[i]
            # print 'New X shape is {}'.format(X.shape)
            self._X = X
        return self._X

    # ======================================
    # y (calculated)
    # ======================================
    @property
    def y(self):
        """
        The :math:`X` weighted properly by the errors from *y_error*
        """
        if self._y is None:
            self._y = self.y_unweighted/self.y_error
        return self._y

    # ======================================
    # y_fit (y from fit)
    # ======================================
    @property
    def y_fit(self):
        """
        Using the result of the linear least squares, the result of :math:`X_{ij}\\beta_i`
        """
        if self._y_fit is None:
            self._y_fit = _np.dot(self.X_unweighted, self.beta)
        return self._y_fit

    # ======================================
    # beta (calculated)
    # ======================================
    @property
    def beta(self):
        """
        The result :math:`\\beta` of the linear least squares
        """
        if self._beta is None:
            # This is the linear least squares matrix formalism
            self._beta = _np.dot(_np.linalg.pinv(self.X) , self.y)
        return self._beta

    # ======================================
    # covar (calculated)
    # ======================================
    @property
    def covar(self):
        """
        The covariance matrix for the result :math:`\\beta`
        """
        if self._covar is None:
            self._covar = _np.linalg.inv(_np.dot(_np.transpose(self.X), self.X))
        return self._covar
    
    # ======================================
    # chisq_red (calculated)
    # ======================================
    @property
    def chisq_red(self):
        """
        The reduced chi-square of the linear least squares
        """
        if self._chisq_red is None:
            self._chisq_red = chisquare(self.y_unweighted.transpose(), _np.dot(self.X_unweighted, self.beta), self.y_error, ddof=3, verbose=False)
        return self._chisq_red



#!/usr/bin/env python3
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import scipy as _sp

from .fill_missing_timestamps import fill_missing_timestamps


def fft(values, freq=None, timestamps=None, fill_missing=False):
    """
    Adds options to :func:`scipy.fftpack.rfft`:

    * *freq* is the frequency the samples were taken at
    * *timestamps* is the time the samples were taken, to help with filling in missing data if *fill_missing* is true
    """
    # ======================================
    # Get frequency
    # ======================================
    if freq is None:
        from .. import qt
        freq = qt.getDouble(title='Fourier Analysis', text='Frequency samples taken at:', min=0, decimals=2, value=1.0)
        freq = freq.input
    
    if fill_missing:
        (t_x, x_filled) = fill_missing_timestamps(timestamps, values)
    else:
        x_filled = values
        
    num_samples = _np.size(x_filled)
    xfft = _sp.fftpack.rfft(x_filled)
    
    factor = freq/num_samples
    num_fft = _np.size(xfft)
    f = factor * _np.linspace(1, num_fft, num_fft)
    
    xpow = _np.abs(xfft*_np.conj(xfft))

    # ======================================
    # No DC term
    # ======================================
    xpow = xpow[1:]
    f = f[1:]

    return (f, xpow)



import jinja2 as _jj
import tempfile as _tempfile
import os as _os
import datetime as _dt
import shutil as _shutil

basedir = '/u1/facet/physics/logbook/data'


# ============================
# Copy files as needed
# ============================
def _copy_file(filepath, fulltime):
    if filepath is None:
        filepath_out = ''
    else:
        filename  = _os.path.basename(filepath)
        root, ext = _os.path.splitext(filename)
        filepath_out = fulltime + ext
        copypath = _os.path.join(basedir, filepath_out)
        _shutil.copyfile(filepath, copypath)

    return filepath_out


def print2elog(author='', title='', text='', link=None, file=None, now=None):
    """
    Prints to the elog given:
    
    * An *author*
    * *title*
    * *link* path to a thumbnail
    * *file* path to a file
    * Time *now*.
    """
    # ============================
    # Get current time
    # ============================
    if now is None:
        now  = _dt.datetime.now()
    fulltime = now.strftime('%Y-%m-%dT%H:%M:%S-00')

    # ============================
    # Copy files
    # ============================
    if not ((link is None) ^ (file is None)):
        link_copied = _copy_file(link, fulltime)
        file_copied = _copy_file(file, fulltime)
    else:
        raise ValueError('Need both file and its thumbnail!')

    # ============================
    # Jinja templating
    # ============================
    loader = _jj.PackageLoader('pytools.facettools', 'resources/templates')
    env = _jj.Environment(loader=loader, trim_blocks=True)
    template = env.get_template('facetelog.xml')
    stream = template.stream(author=author, title=title, text=text, link=link_copied, file=file_copied, now=now)

    # ============================
    # Write xml
    # ============================
    with _tempfile.TemporaryDirectory() as dirname:
        filename = '{}.xml'.format(fulltime)
        filepath = _os.path.join(dirname, filename)

        with open(filepath, 'w+') as fid:
            # stream.disable_buffering()
            stream.dump(fid)

        finalpath = _os.path.join(basedir, filename)
        # _shutil.copyfile(filepath, 'new.xml')
        _shutil.copyfile(filepath, finalpath)



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    pass


class BeamBase(object):
    def __init__(self, nb0, E=20.35, dE=0.01):
        self._nb0 = nb0
        self._E   = E
        self._dE  = dE

    @property
    def nb0(self):
        """
        On-axis beam density.
        """
        return self._nb0

    @property
    def E(self):
        """
        Beam energy in GeV.
        """
        return self._E

    @property
    def dE(self):
        """
        Beam energy spread.
        """
        return self._dE


class RoundBeam(BeamBase):
    def __init__(self, nb0, s_r, E=20.35, dE=0.01):
        super().__init__(nb0=nb0, E=E, dE=dE)
        self._s_r = s_r

    @property
    def s_r(self):
        return self._s_r



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.image as _mplim
    import matplotlib.cm as _cm
    import numpy as _np

from .colorbar import colorbar as _cb
from .setup_axes import setup_axes as _setup_axes
from .SciImage import SciImage as _SI


def NonUniformImage(x, y, z, ax=None, fig=None, cmap=None, alpha=None, scalex=True, scaley=True, add_cbar=True, **kwargs):
    """
    Plots a set of coordinates where:

    * *x* and *y* are 1-D ndarrays of lengths N and M, respectively, specifying pixel centers
    * *z* is an (M, N) ndarray or masked array of values to be colormapped, or a (M, N, 3) RGB array, or a (M, N, 4) RGBA array.

    *\*\*kwargs* can contain keywords:
    
    * *cmap* for set the colormap
    * *alpha* to set transparency
    * *scalex* to set the x limits to available data
    * *scaley* to set the y limits to available data

    Returns class :class:`matplotlib.image.NonUniformImage`.
    """
    if ax is None and fig is None:
        fig, ax = _setup_axes()
    elif ax is None:
        ax = fig.gca()
    elif fig is None:
        fig = ax.get_figure()

    im = _mplim.NonUniformImage(ax, **kwargs)

    vmin = kwargs.pop('vmin', _np.min(z))
    vmax = kwargs.pop('vmax', _np.max(z))
    im.set_clim(vmin=vmin, vmax=vmax)

    if cmap is not None:
        im.set_cmap(cmap)

    m = _cm.ScalarMappable(cmap=im.get_cmap())
    m.set_array(z)

    if add_cbar:
        cax, cb = _cb(ax=ax, im=m, fig=fig)

    if alpha is not None:
        im.set_alpha(alpha)

    im.set_data(x, y, z)
    ax.images.append(im)

    if scalex:
        xmin = min(x)
        xmax = max(x)
        ax.set_xlim(xmin, xmax)

    if scaley:
        ymin = min(y)
        ymax = max(y)
        ax.set_ylim(ymin, ymax)

    return _SI(im=im, cb=cb, cax=cax)



class SciImage(object):
    def __init__(self, im, cb, cax):
        self._im = im
        self._cb = cb
        self._cax = cax

    @property
    def im(self):
        return self._im

    @property
    def cb(self):
        return self._cb

    @property
    def cax(self):
        return self._cax



__all__ = [
    'Imshow_Slider',
    'Imshow_Slider_Array',
    'NonUniformImage',
    'NonUniformImage_axes',
    'RectangleSelector',
    'addlabel',
    'axesfontsize',
    'colorbar',
    'contour',
    'figure',
    'hist',
    'hist2d',
    'imshow',
    'imshow_batch',
    'latexfig',
    'less_labels',
    'pcolor_axes',
    'plot',
    'plot_featured',
    'quiver',
    'rgb2gray',
    'savefig',
    'scaled_figsize',
    'setup_axes',
    'setup_figure',
    'showfig',
    ]
__all__.sort()

import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    from .cmaps import *  # noqa
    __all__.append('viridis')
    __all__.sort()

from .Imshow_Slider import Imshow_Slider
from .Imshow_Slider_Array import Imshow_Slider_Array
from .NonUniformImage import NonUniformImage
from .NonUniformImage_axes import NonUniformImage_axes
from .RectangleSelector import RectangleSelector
from .addlabel import addlabel
from .axesfontsize import axesfontsize
from .colorbar import colorbar
from .figure import figure
from .hist import hist
from .hist2d import hist2d
from .imshow import contour
from .imshow import imshow
from .imshow import quiver
from .imshow import scaled_figsize
from .imshow_batch import imshow_batch
from .latexfig import latexfig
from .less_labels import less_labels
from .pcolor_axes import pcolor_axes
from .plot import plot
from .plot_featured import plot_featured
from .rgb2gray import rgb2gray
from .savefig import savefig
from .setup_axes import setup_axes
from .setup_figure import setup_figure
from .showfig import showfig



__all__ = [
    'Beam',
    'Ions1D',
    'Ions2D',
    'Match',
    'MatchPlasma',
    'Plasma',
    ]
__all__.sort()

from .ions1d import Ions1D
from .ions2d import Ions2D
from .match import *            # noqa
from .plasma import *           # noqa
from .beam import RoundBeam



"""
The :mod:`numpy <scisalt.numpy>` module contains a few convenience functions mostly designed to make evaluating functions easier for plotting.
"""
__all__ = [
    'frexp10',
    'gaussian',
    'linspaceborders',
    'linspacestep',
    'piecewise',
    ]

__all__.sort()

from .frexp10 import *
from .functions import gaussian
from .linspaceborders import *
from .linspacestep import *
from .piecewise import piecewise



from PyQt4 import QtGui
from PyQt4 import QtCore
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as _FigureCanvas
from matplotlib.backends.backend_qt4 import NavigationToolbar2QT as _NavigationToolbar
import matplotlib as _mpl
import numpy as _np

from .Rectangle import Rectangle

import pdb
import traceback

import logging
loggerlevel = logging.DEBUG
logger = logging.getLogger(__name__)

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8

    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)


class Slider_and_Text(QtGui.QWidget):
    valueChanged = QtCore.pyqtSignal(int)
    sliderReleased = QtCore.pyqtSignal(int)

    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self)
        self.setMaximumHeight(40)
        # Enable tracking by default
        self._tracking = True
        self.hLayout   = QtGui.QHBoxLayout()
        self.slider    = QtGui.QSlider()

        self.leftbutton = QtGui.QPushButton()
        self.leftbutton.setText("<")
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leftbutton.sizePolicy().hasHeightForWidth())
        #  self.leftbutton.setSizePolicy(sizePolicy)
        self.leftbutton.clicked.connect(self._subone)

        self.rightbutton = QtGui.QPushButton()
        self.rightbutton.setText(">")
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rightbutton.sizePolicy().hasHeightForWidth())
        #  self.rightbutton.setSizePolicy(sizePolicy)
        self.rightbutton.clicked.connect(self._addone)

        self.v = QtGui.QIntValidator()
        self.box = QtGui.QLineEdit()
        self.box.setValidator(self.v)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.box.sizePolicy().hasHeightForWidth())
        #  self.box.setSizePolicy(sizePolicy)

        self.hLayout.addWidget(self.leftbutton)
        self.hLayout.addWidget(self.slider)
        self.hLayout.addWidget(self.box)
        self.hLayout.addWidget(self.rightbutton)
        self.setLayout(self.hLayout)
        
        self.slider.valueChanged.connect(self._sliderChanged)
        self.box.editingFinished.connect(self._textChanged)
        self.setOrientation(QtCore.Qt.Horizontal)

        # Connect release so tracking works as expected
        self.slider.sliderReleased.connect(self._sliderReleased)

    def _addone(self):
        self.value = self.value + 1
        self.valueChanged.emit(self.value)

    def _subone(self):
        self.value = self.value - 1
        self.valueChanged.emit(self.value)

    def _sliderReleased(self):
        print('Released')
        self.sliderReleased.emit(self.slider.value)

    def setTracking(self, val):
        print('Tracking set to {}'.format(val))
        self._tracking = val

    def setMaximum(self, val):
        self.slider.setMaximum(val)
        self.v.setRange(self.slider.minimum(), self.slider.maximum())
        self.box.setValidator(self.v)

    def setMinimum(self, val):
        self.slider.setMinimum(val)
        self.v.setRange(self.slider.minimum(), self.slider.maximum())
        self.box.setValidator(self.v)

    def _sliderChanged(self, val):
        self.box.setText(str(val))
        if self._tracking:
            try:
                self.slider.sliderReleased.disconnect()
            except:
                pass
            self.valueChanged.emit(val)
        else:
            try:
                self.slider.sliderReleased.disconnect()
            except:
                pass
            self.slider.sliderReleased.connect(self._sliderChanged_notracking)

    def _sliderChanged_notracking(self):
        val = self.slider.value()
        # print('Value to be emitted is {}'.format(val))
        self.valueChanged.emit(val)

    def _textChanged(self):
        val = self.box.text()
        self.slider.setValue(int(val))
        self._sliderChanged_notracking()

    def setOrientation(self, *args, **kwargs):
        self.slider.setOrientation(*args, **kwargs)

    def _getValue(self):
        return self.slider.value()

    def _setValue(self, val):
        self.slider.setValue(val)
        self.box.setText(str(val))
    value = property(_getValue, _setValue)

    def setValue(self, val):
        self.slider.setValue(val)
        self.box.setText(str(val))
        # self.valueChanged.emit(val)


class Mpl_Plot(_FigureCanvas):
    def __init__(self, parent=None):
        # Initialize things
        self.fig = _mpl.figure.Figure()
        _FigureCanvas.__init__(self, self.fig)
        _FigureCanvas.setSizePolicy(self, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        _FigureCanvas.updateGeometry(self)

        # Create axes
        self.ax = self.fig.add_subplot(111)

    def plot(self, *args, **kwargs):
        self.ax.clear()
        self.ax.plot(*args, **kwargs)
        self.ax.ticklabel_format(style='sci', scilimits=(0, 0), axis='y')
        self.ax.figure.canvas.draw()


class Mpl_Image(QtGui.QWidget):
    # Signal for when the rectangle is changed
    rectChanged = QtCore.pyqtSignal(Rectangle)

    def __init__(self, parent=None, rectbool = True, toolbarbool=False, image=None):
        # Initialize things
        QtGui.QWidget.__init__(self)
        self.rectbool  = rectbool
        self._clim_min = 0
        self._clim_max = 3600

        self._pressed = False

        # Add a vertical layout
        self.vLayout = QtGui.QVBoxLayout()

        # Add a figure
        self.fig = _mpl.figure.Figure()

        # Add a canvas containing the fig
        self.canvas = _FigureCanvas(self.fig)
        _FigureCanvas.setSizePolicy(self.canvas, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        _FigureCanvas.updateGeometry(self.canvas)

        # Setup the layout
        if toolbarbool:
            self.toolbar = _NavigationToolbar(self.canvas, self)
            self.toolbar.setMaximumHeight(20)
            self.vLayout.addWidget(self.toolbar)
        self.vLayout.addWidget(self.canvas)
        self.setLayout(self.vLayout)

        # Create axes
        self.ax = self.fig.add_subplot(111)

        # Include rectangle functionality
        if rectbool:
            self.fig.canvas.mpl_connect('button_press_event', self.on_press)
            self.fig.canvas.mpl_connect('button_release_event', self.on_release)
            self.Rectangle = Rectangle(
                x      = -10     ,
                y      = 0       ,
                width  = 0       ,
                height = 3       ,
                axes   = self.ax
                )

        # Add image
        self.image = image

    def _get_img(self):
        return self._image

    def _set_img(self, image):
        self.ax.clear()
        self._image = image
        if image is not None:
            self._imgplot = self.ax.imshow(image, interpolation='none')
            if self.rectbool:
                self.ax.add_patch(self.Rectangle.get_rect())
            # imagemax = _np.max(_np.max(image))
            self.set_clim(self._clim_min, self._clim_max)
    image = property(_get_img, _set_img)

    def set_clim(self, clim_min, clim_max):
        if self.image is not None:
            self._clim_min = clim_min
            self._clim_max = clim_max
            self._imgplot.set_clim(clim_min, clim_max)
            self.ax.figure.canvas.draw()

    def on_press(self, event):
        if self.toolbar._active is None:
            self._pressed = True
            self.x0 = event.xdata
            self.y0 = event.ydata
            logger.log(level=loggerlevel, msg='Pressed: x0: {}, y0: {}'.format(self.x0, self.y0))

    def on_release(self, event):
        if self._pressed:
            self._pressed = False
            print('release')
            self.x1 = event.xdata
            self.y1 = event.ydata
            width   = self.x1 - self.x0
            height  = self.y1 - self.y0

            logger.log(level=loggerlevel, msg='Released: x0: {}, y0: {}, x1: {}, y1: {}, width: {}, height: {}'.format(
                self.x0 ,
                self.y0 ,
                self.x1 ,
                self.y1 ,
                width   ,
                height
                )
                )

            self.Rectangle.set_xy((self.x0, self.y0))
            self.Rectangle.set_width(width)
            self.Rectangle.set_height(height)
            self.ax.figure.canvas.draw()

            self.rectChanged.emit(self.Rectangle)
            # print(self.rect)

    def zoom_rect(self, border=None, border_px=None):
        # ======================================
        # Get x coordinates
        # ======================================
        x0 = self.Rectangle.get_x()
        width = self.Rectangle.get_width()
        x1 = x0+width

        # ======================================
        # Get y coordinates
        # ======================================
        y0 = self.Rectangle.get_y()
        height = self.Rectangle.get_height()
        y1 = y0+height

        # ======================================
        # Validate borders
        # ======================================
        if (border_px is None) and (border is not None):
            xborder = border[0]*width
            yborder = border[1]*height
        elif (border_px is not None) and (border is None):
            xborder = border_px[0]
            yborder = border_px[1]
        elif (border_px is None) and (border is None):
            raise IOError('No border info specified!')
        elif (border_px is not None) and (border is not None):
            raise IOError('Too much border info specified, both border_px and border!')
        else:
            raise IOError('End of the line!')

        # ======================================
        # Add borders
        # ======================================
        x0 = x0 - xborder
        x1 = x1 + xborder
        y0 = y0 - yborder
        y1 = y1 + yborder

        # ======================================
        # Validate coordinates to prevent
        # unPythonic crash
        # ======================================
        if not ((0 <= x0 and x0 <= self.image.shape[1]) and (0 <= x1 and x1 <= self.image.shape[1])):
            print('X issue')
            print('Requested: x=({}, {})'.format(x0, x1))
            x0 = 0
            x1 = self.image.shape[1]
        if not ((0 <= y0 and y0 <= self.image.shape[0]) and (0 <= y1 and y1 <= self.image.shape[0])):
            print('y issue')
            print('Requested: y=({}, {})'.format(y0, y1))
            y0 = 0
            y1 = self.image.shape[0]

        # ======================================
        # Set viewable area
        # ======================================
        self.ax.set_xlim(x0, x1)
        self.ax.set_ylim(y0, y1)

        # ======================================
        # Redraw canvas to show updates
        # ======================================
        self.ax.figure.canvas.draw()


class Mpl_Image_Plus_Slider(QtGui.QWidget):
    # def __init__(self, parent=None, **kwargs):
    def __init__(self, parent=None, **kwargs):
        # Initialize self as a widget
        QtGui.QWidget.__init__(self, parent)

        # Add a vertical layout with parent self
        self.vLayout = QtGui.QVBoxLayout(self)
        self.vLayout.setObjectName(_fromUtf8("vLayout"))

        # Add an Mpl_Image widget to vLayout,
        # save it to self._img
        # Pass arguments through to Mpl_Image.
        self._img = Mpl_Image(parent=parent, toolbarbool=True, **kwargs)
        self._img.setObjectName(_fromUtf8("_img"))
        self.vLayout.addWidget(self._img)

        # Add a slider to vLayout,
        # save it to self.max_slider
        # self.max_slider = QtGui.QSlider(self)
        self.max_slider = Slider_and_Text(self)
        self.max_slider.setObjectName(_fromUtf8("max_slider"))
        self.max_slider.setOrientation(QtCore.Qt.Horizontal)
        self.vLayout.addWidget(self.max_slider)

        # Setup slider to work with _img's clims
        self.max_slider.valueChanged.connect(lambda val: self.set_clim(0, val))

    def _get_image(self):
        return self._img.image

    def _set_image(self, image):
        self._img.image = image
        maximage = _np.max(_np.max(image))
        self.max_slider.setMaximum(maximage)
    image = property(_get_image, _set_image)

    def _get_ax(self):
        return self._img.ax
    ax = property(_get_ax)

    def _get_Rectangle(self):
        return self._img.Rectangle
    #  def _set_rect(self, rect):
    #          self._img.rect(rect)
    Rectangle = property(_get_Rectangle)

    def zoom_rect(self, border=None, border_px=None):
        self._img.zoom_rect(border, border_px)

    def set_clim(self, *args, **kwargs):
        self._img.set_clim(*args, **kwargs)

    def setSliderValue(self, val):
        self.max_slider.setValue(val)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


def NonUniformImage_axes(img):
    """
    Returns axes *x, y* for a given image *img* to be used with :func:`scisalt.matplotlib.NonUniformImage`.
    """
    xmin = 0
    xmax = img.shape[1]-1
    ymin = 0
    ymax = img.shape[0]-1
    x = _np.linspace(xmin, xmax, img.shape[1])
    y = _np.linspace(ymin, ymax, img.shape[0])
    return x, y



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def frexp10(x):
    """
    Finds the *mantissa* :math:`m` and *exponent* :math:`e` of a number *x* such that :math:`x = m 10^e`

    Returns the list :code:`(mantissa, exponent)`.
    """
    expon = _np.int(_np.floor(_np.log10(_np.abs(x))))
    mant = x/_np.power(10, expon)
    return (mant, expon)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import matplotlib.pyplot as _plt

from .addlabel import addlabel as _addlabel


def hist(x, bins=10, labels=None, aspect="auto", plot=True, ax=None, range=None):
    """
    Creates a histogram of data *x* with a *bins*, *labels* = :code:`[title, xlabel, ylabel]`.
    """

    h, edge = _np.histogram(x, bins=bins, range=range)
    
    mids = edge + (edge[1]-edge[0])/2
    mids = mids[:-1]

    if plot:
        if ax is None:
            _plt.hist(x, bins=bins, range=range)
        else:
            ax.hist(x, bins=bins, range=range)

        if labels is not None:
            _addlabel(labels[0], labels[1], labels[2])

    return h, mids



import logging as _logging
import inspect as _inspect
__all__ = ['mylogger', 'log']


def mylogger(name=None, filename=None, indent_offset=7, level=_logging.DEBUG, stream_level=_logging.WARN, file_level=_logging.INFO):
    """
    Sets up logging to *filename*.debug.log, *filename*.log, and the terminal. *indent_offset* attempts to line up the lowest indent level to 0. Custom levels:

    * *level*: Parent logging level.
    * *stream_level*: Logging level for console stream.
    * *file_level*: Logging level for general file log.
    """
    if name is not None:
        logger = _logging.getLogger(name)
    else:
        logger = _logging.getLogger()

    logger.setLevel(level)

    fmtr         = IndentFormatter(indent_offset=indent_offset)
    fmtr_msgonly = IndentFormatter('%(funcName)s:%(lineno)d: %(message)s')

    ch = _logging.StreamHandler()
    ch.setLevel(stream_level)
    ch.setFormatter(fmtr_msgonly)
    logger.addHandler(ch)

    if filename is not None:
        debugh = _logging.FileHandler(filename='{}_debug.log'.format(filename), mode='w')
        debugh.setLevel(_logging.DEBUG)
        debugh.setFormatter(fmtr_msgonly)
        logger.addHandler(debugh)

        fh = _logging.FileHandler(filename='{}.log'.format(filename), mode='w')
        fh.setLevel(file_level)
        fh.setFormatter(fmtr)
        logger.addHandler(fh)

    return logger


def log(logger, level):
    def log(msg):
        return logger.log(level=level, msg=msg)
    return log


class IndentFormatter(_logging.Formatter):
    def __init__( self, fmt=None, datefmt=None, indent_offset=6):
        if fmt is None:
            fmt = '%(indent)s==========================================================\n%(indent)s%(levelname)s - %(name)s:%(funcName)s:%(lineno)d\n%(indent)s%(message)s'

        super(IndentFormatter, self).__init__(fmt=fmt, datefmt=datefmt)
        self.baseline = len(_inspect.stack()) + indent_offset

    def format( self, rec ):
        stack = _inspect.stack()
        stackdepth = len(stack)
        stackdiff = stackdepth - self.baseline
        rec.indent = '\t' * stackdiff
        #  rec.function = stack[8][3]
        out = _logging.Formatter.format(self, rec)
        del rec.indent
        #  del rec.function
        return out



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


def px_to_units(xpx, ypx):
    return xpx, ypx


def pcolor_axes(array, px_to_units=px_to_units):
    """
    Return axes :code:`x, y` for *array* to be used with :func:`matplotlib.pyplot.color`.

    *px_to_units* is a function to convert pixels to units. By default, returns pixels.
    """
    # ======================================
    # Coords need to be +1 larger than array
    # ======================================
    x_size = array.shape[0]+1
    y_size = array.shape[1]+1

    x = _np.empty((x_size, y_size))
    y = _np.empty((x_size, y_size))

    for i in range(x_size):
        for j in range(y_size):
            x[i, j], y[i, j] = px_to_units(i-0.5, j-0.5)

    return x, y



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def keys(f):
    """
    .. versionadded:: 1.4

    Returns an array of strings of the keys like Python 2 used to do.
    """
    return [key for key in f.keys()]


def get(f, key, default=None):
    """
    .. versionadded:: 1.4

    Gets an array from datasets.
    """

    if key in f.keys():
        val = f[key].value

        if default is None:
            return val
        else:
            if _np.shape(val) == _np.shape(default):
                return val

    return default



# from .linspacestep import linspacestep as _linspacestep
import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def linspaceborders(array):
    """
    Given a 1-D *array*, generate a new array with numbers between the numbers of the input *array*. Adds elements to the left and right sides to get the exterior border as well.

    Returns a :class:`numpy.ndarray`.
    """
    # Get and set left side
    dela = array[1] - array[0]
    new_arr = _np.array([array[0] - dela / 2])

    # Add a point to the right side so the for loop works
    delb = array[-1] - array[-2]
    array = _np.append(array, array[-1] + delb)
    for i, val in enumerate(array):
        try:
            avg = (array[i] + array[i + 1]) / 2
            new_arr = _np.append(new_arr, avg)
        except:
            pass
    # start = array[0] - dela / 2
    # end = array[-1] + dela / 2
    # return _linspacestep(start, end, dela)
    return new_arr



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import matplotlib as mpl


class Rectangle(object):
    def __init__(self, x, y, width, height, axes=None, alpha=0.5, fill=True):
        self._axes = axes
        self._rect = mpl.patches.Rectangle((x, y), width, height, facecolor='w', edgecolor='r', alpha=alpha, fill=fill, axes=self._axes)

    @property
    def patch(self):
        return self._rect

    @property
    def _sorted_x(self):
        x_a = self._rect.get_x()
        x_b = x_a + self._rect.get_width()
        return _np.sort([x_a, x_b])

    @property
    def x0(self):
        """
        The smaller x coordinate.
        """
        return self._sorted_x[0]

    @property
    def x1(self):
        """
        The larger x coordinate.
        """
        return self._sorted_x[1]
    
    def get_x(self):
        return self.x0

    @property
    def _sorted_y(self):
        y_a = self._rect.get_y()
        y_b = y_a + self._rect.get_height()
        return _np.sort([y_a, y_b])

    @property
    def y0(self):
        """
        The smaller x coordinate.
        """
        return self._sorted_y[0]

    @property
    def y1(self):
        """
        The larger x coordinate.
        """
        return self._sorted_y[1]

    def get_y(self):
        return self.y0

    def get_xy(self):
        return (self.get_x(), self.get_y())

    def get_width(self):
        return self.x1-self.x0

    def get_height(self):
        return self.y1-self.y0

    def set_xy(self, value):
        return self._rect.set_xy(value)

    def set_width(self, value):
        return self._rect.set_width(value)

    def set_height(self, value):
        return self._rect.set_height(value)



# import matplotlib.pyplot as _plt

import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.widgets as _wdg
    import matplotlib as _mpl
    import numpy as _np
    import matplotlib.pyplot as _plt

from .setup_figure import setup_figure   # noqa
from .imshow import imshow
from .imshow import scaled_figsize
from .colorbar import colorbar

import pdb as pdb


class Imshow_Slider_Array(object):
    """
    .. versionchanged:: 1.5
       Name changed, colorbar options added, *p* changed to :class:`AxesImage <scisalt.matplotlib.Imshow_Slider.AxesImage>`.

    Convenience class for viewing images.
    
    Plots *image* to a to an instance of :class:`matplotlib.axis.imshow(**kwargs)`, with sliders for controlling bounds, with *\*\*kwargs* passed through to :meth:`matplotlib.axes.Axes.imshow`.

    *usecbar* determines if a colorbar will be used. Color bars can slow down the viewer significantly.
    """
    def __init__(self, images, usecbar=False, **kwargs):
        # ======================================
        # Save input info
        # ======================================
        self._images    = images
        self._num_imgs  = len(images)
        self._image_ind = 0
        self._kwargs    = kwargs
        self._usecbar   = usecbar

        # ======================================
        # Create figure
        # ======================================
        # self.fig, self.gs = setup_figure(20, 10, figsize=scaled_figsize(self.image))
        self.fig, self.gs = setup_figure(20, 10)
        self._ax_img = self.fig.add_subplot(self.gs[0:-4, :])
        self.ax_min = self.fig.add_subplot(self.gs[-3, 1:-1])
        self.ax_max = self.fig.add_subplot(self.gs[-2, 1:-1])
        self.ax_img = self.fig.add_subplot(self.gs[-1, 1:-1])

        self._reset(**kwargs)

        _plt.connect('key_press_event', self._keypress)

    def _keypress(self, event):
        self._event = event
        newval = None
        if event.key == 'right':
            newval = self.imageslider.val + 1
            if newval > self.num_imgs:
                newval = self.num_imgs
        elif event.key == 'left':
            newval = self.imageslider.val - 1
            if newval < 0:
                newval = 0

        if newval is not None:
            self.imageslider.set_val(newval)

        # if event.key in ['A', 'a']:
        #     self.RectangleSelector.set_active(not self.RectangleSelector.active)
        # elif event.key in ['D', 'd']:
        #     try:
        #         self._rect.remove()
        #         _plt.draw()

        #         self._eclick          = None
        #         self._erelease        = None
        #         self._selfunc_results = None
        #     except:
        #         pass

    def _fix_axes(self):
        shape_x, shape_y = self.image.shape
        ratio = shape_x / shape_y
        figsize = self._kwargs.get('figsize', _mpl.rcParams['figure.figsize'])
        figsize_ratio = figsize[0]/figsize[1]
        if ratio > figsize_ratio:
            x_lim = [0, shape_x]
            y_lim = shape_x / figsize_ratio
            y_lim = [(shape_y-y_lim)/2, (shape_y+y_lim)/2]
        else:
            x_lim = shape_y * figsize_ratio
            x_lim = [(shape_x-x_lim)/2, (shape_x+x_lim)/2]
            y_lim = [0, shape_y]

        self.ax.set_xlim(x_lim)
        self.ax.set_ylim(y_lim)

    def _reset(self, **kwargs):
        # ======================================
        # Strip kwargs for vmin, vmax
        # in order to set sliders correctly
        # ======================================
        minslide = kwargs.get('vmin', self.imgmin)
        maxslide = kwargs.get('vmax', self.imgmax)

        # ======================================
        # Slider Logic
        # ======================================
        if self.imgmin > 0:
            slidermin = kwargs.pop('smin', 0)
        else:
            slidermin = kwargs.pop('smin', self.imgmin)

        if self.imgmax < 0:
            slidermax = kwargs.pop('smax', 0)
        else:
            slidermax = kwargs.pop('smax', self.imgmax)

        # ======================================
        # Imshow
        # ======================================
        # self._AxesImage = self.ax.imshow(self.image, **kwargs)
        self._AxesImage = imshow(self.image, ax=self.ax, add_cbar=False, **kwargs)
        self._cax, self._cb = colorbar(self.ax, self._AxesImage)
        
        self._fix_axes()

        # ======================================
        # Add minimum slider
        # ======================================
        self.minslider = _wdg.Slider(self.ax_min, 'Min', slidermin, slidermax, minslide)

        # ======================================
        # Add maximum slider
        # ======================================
        self.maxslider = _wdg.Slider(self.ax_max, 'Max', slidermin, slidermax, maxslide, slidermin=self.minslider)
        self.minslider.slidermax = self.maxslider

        self.minslider.on_changed(self._update_clim)
        self.maxslider.on_changed(self._update_clim)

        # ======================================
        # Add image slider
        # ======================================
        self.imageslider = _wdg.Slider(ax=self.ax_img, label='Image', valmin=1, valmax=self.num_imgs, valinit=0, valfmt=u'%d')
        self.imageslider.on_changed(self._update_image)

        # if self.usecbar:
        #     self.fig.colorbar(self.AxesImage, ax=self.ax, use_gridspec=True)

        # self.fig.tight_layout()

    def _update_image(self, value):
        ind = _np.int(_np.round(value)-1)
        self._image_ind = ind

        # ======================================
        # Get old axes settings
        # ======================================
        xlim = self.ax.get_xlim()
        ylim = self.ax.get_ylim()

        # ======================================
        # Clear old axes
        # ======================================
        self.ax.cla()
        self._cax.cla()

        # ======================================
        # Plot new data
        # ======================================
        self._AxesImage = imshow(self.image, ax=self.ax, add_cbar=False, **self._kwargs)
        self._cb = _plt.colorbar(self._AxesImage, cax=self._cax)

        # ======================================
        # Fix the axes
        # ======================================
        self._fix_axes()

        # ======================================
        # Keep the colorbar to previous settings
        # ======================================
        self._update_clim(0)

        # ======================================
        # Get old axes settings
        # ======================================
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)

        
    def set_cmap(self, cmap):
        """
        Sets color map to *cmap*.
        """
        self.AxesImage.set_cmap(cmap)

    @property
    def num_imgs(self):
        return self._num_imgs

    @property
    def AxesImage(self):
        """
        The :class:`matplotlib.image.AxesImage` from :meth:`matplotlib.axes.Axes.imshow`.
        """
        return self._AxesImage

    @property
    def ax(self):
        """
        The :class:`matplotlib.axes.Axes` used for :meth:`matplotlib.axes.Axes.imshow`.
        """
        return self._ax_img

    # ======================================
    # Get max of image
    # ======================================
    @property
    def imgmax(self):
        """
        Highest value of input image.
        """
        if not hasattr(self, '_imgmax'):
            imgmax = _np.max(self.images[0])
            for img in self.images:
                imax = _np.max(img)
                if imax > imgmax:
                    imgmax = imax

            self._imgmax = imgmax

        return self._imgmax

    # ======================================
    # Get min of image
    # ======================================
    @property
    def imgmin(self):
        """
        Lowest value of input image.
        """
        if not hasattr(self, '_imgmin'):
            imgmin = _np.min(self.images[0])
            for img in self.images:
                imin = _np.min(img)
                if imin > imgmin:
                    imgmin = imin

            self._imgmin = imgmin
        return _np.min(self.image)

    # ======================================
    # Update the clims
    # ======================================
    def _update_clim(self, val):
        cmin = self.minslider.val
        cmax = self.maxslider.val
        # print('Cmin: {}, Cmax: {}'.format(cmin, cmax))
        self.AxesImage.set_clim(cmin, cmax)

    # ======================================
    # Easily get and set slider
    # ======================================
    @property
    def clim_min(self):
        """
        Slider value for minimum
        """
        return self.minslider.val

    @clim_min.setter
    def clim_min(self, val):
        self.minslider.set_val(val)

    @property
    def clim_max(self):
        """
        Slider value for maximum
        """
        return self.maxslider.val

    @clim_max.setter
    def clim_max(self, val):
        self.maxslider.set_val(val)

    @property
    def images(self):
        """
        The array of images.
        """
        return self._images

    @property
    def image(self):
        """
        The image loaded.
        """
        return self._images[self._image_ind]



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import matplotlib.widgets as _mw
    import matplotlib.pyplot as _plt
    import matplotlib.patches as _mp

__all__ = ['RectangleSelector']


class RectangleSelector(object):
    """
    .. versionadded:: 1.2

    Add rectangle selection to an already-existing axis *ax*. *\*args* and *\*\*kwargs* pass through to :class:`matplotlib.widgets.RectangleSelector`.

    Use key *A* or *a* to toggle whether the rectangle is active or not.

    *verbose* controls whether selections are printed to the terminal.
    """
    def __init__(self, ax, *args, selfunc=None, parent=None, verbose=False, **kwargs):
        # ======================================
        # Store things to class
        # ======================================
        self._parent          = parent
        self._selfunc         = selfunc
        self._ax              = ax
        self.verbose          = verbose
        self._selfunc_results = None
        self._x0 = 0
        self._x1 = 1
        self._y0 = 0
        self._y1 = 1

        # ======================================
        # Add rectangle selector
        # ======================================
        self._RectangleSelector = _mw.RectangleSelector(ax, self._onselect, *args, **kwargs)
        _plt.connect('key_press_event', self._toggle)

    def _onselect(self, eclick, erelease):
        # ======================================
        # Occurs on release
        # ======================================
        self._eclick = eclick
        self._erelease = erelease
        if self._verbose:
            print('eclick (x, y):\t\t({}, {})'.format(eclick.xdata, eclick.ydata))
            print('erelease (x, y):\t({}, {})'.format(erelease.xdata, erelease.ydata))

        self._x0 = min([self.erelease.xdata, self.eclick.xdata])
        self._x1 = max([self.erelease.xdata, self.eclick.xdata])
        self._y0 = min([self.erelease.ydata, self.eclick.ydata])
        self._y1 = max([self.erelease.ydata, self.eclick.ydata])

        self._set_rect()

        self._selfunc_results = self.selfunc(self)

    def _set_rect(self):
        try:
            self._rect.remove()
        except:
            pass

        self._rect = self._ax.add_patch(
            _mp.Rectangle(
                xy     = (self.x0, self.y0),
                width  = self.width,
                height = self.height,
                ec     = 'r',
                fc     = 'none'
                )
            )

        _plt.draw()

    def _toggle(self, event):
        if event.key in ['A', 'a']:
            self.RectangleSelector.set_active(not self.RectangleSelector.active)
        elif event.key in ['D', 'd']:
            try:
                self._rect.remove()
                _plt.draw()

                self._eclick          = None
                self._erelease        = None
                self._selfunc_results = None
            except:
                pass

    @property
    def parent(self):
        return self._parent

    @property
    def verbose(self):
        """
        Determines whether rectangle coordinates are printed to the terminal on selection.
        """
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    @property
    def RectangleSelector(self):
        """
        The instance of :class:`matplotlib.widgets.RectangleSelector`.
        """
        return self._RectangleSelector

    @property
    def ax(self):
        """
        The axis used.
        """
        return self._ax

    @property
    def selfunc(self):
        """
        A placeholder for the function called on each mouse release.
        """
        return self._selfunc

    @property
    def selfunc_results(self):
        """
        The results of :func:`selfunc(instance) <scisalt.matplotlib.RectangleSelector.selfunc>` where *instance* is this class.
        """
        if self.selfunc is not None:
            return self.selfunc(self)

    @property
    def eclick(self):
        """
        The starting mouse click from :class:`RectangleSelector <scisalt.matplotlib.RectangleSelector.RectangleSelector>`.
        """
        if self._eclick is None:
            raise IOError('No area was selected')
        else:
            return self._eclick

    @property
    def erelease(self):
        """
        The ending mouse click from :class:`RectangleSelector <scisalt.matplotlib.RectangleSelector.RectangleSelector>`.
        """
        if self._erelease is None:
            raise IOError('No area was selected')
        else:

            return self._erelease

    @property
    def x0(self):
        """
        Minimum x coordinate of rectangle.
        """
        return self._x0

    @x0.setter
    def x0(self, val):
        self._x0 = val
        self._set_rect()

    @property
    def x1(self):
        """
        Maximum x coordinate of rectangle.
        """
        return self._x1

    @x1.setter
    def x1(self, val):
        self._x1 = val
        self._set_rect()

    @property
    def y0(self):
        """
        Minimum y coordinate of rectangle.
        """
        return self._y0

    @y0.setter
    def y0(self, val):
        self._y0 = val
        self._set_rect()

    @property
    def y1(self):
        """
        Maximum y coordinate of rectangle.
        """
        return self._y1

    @y1.setter
    def y1(self, val):
        self._y1 = val
        self._set_rect()

    @property
    def xslice(self):
        return [self.x0, self.x1]

    @xslice.setter
    def xslice(self, val):
        self._x0, self._x1 = val
        self._set_rect()

    @property
    def yslice(self):
        return [self.y0, self.y1]

    @yslice.setter
    def yslice(self, val):
        self._y0, self._y1 = val
        self._set_rect()

    @property
    def width(self):
        """
        Width of rectangle.
        """
        return self.x1-self.x0

    @property
    def height(self):
        """
        Height of rectangle.
        """
        return self.y1-self.y0



# Author: Joel Frederico
"""
This project has slowly grown from frequently-used code snippets. The main objective is to create a collection of interconnected methods frequently needed to visualize and analyze data using `NumPy <http://www.numpy.org/>`_, `SciPy <http://www.scipy.org/>`_, `Matplotlib <http://matplotlib.org/>`_, and `PyQt4 <http://www.riverbankcomputing.com/software/pyqt/download>`_.
"""
__version__ = '1.8.0'
from . import accelphys
from . import facettools
from . import logging
from . import matplotlib
from . import numpy
from . import scipy
from . import utils
from . import PWFA


# ============================
# Check if master, dev, or tag
# ============================
# import git as _git
# import os as _os
# 
# _path = _os.path.dirname(_os.path.dirname(__file__))
# 
# def _test_git():
#     try:
#         _repo = _git.Repo(_path)
#         _hexsha = _repo.head.object.hexsha
# 
#         if _repo.is_dirty():
#             import warnings as _warnings
#             _warnings.warn('SciSalt repo is dirty, version not well-defined.', category=SyntaxWarning, stacklevel=3)
#     
#         for _tag in _repo.tags:
#             if _tag.object.hexsha == _hexsha:
#                 return
#     
#         import warnings as _warnings
#         _warnings.warn('SciSalt not currently on a tag, version not well-defined. Head: {}'.format(_repo.head.ref.name), category=SyntaxWarning, stacklevel=3)
#     except _git.InvalidGitRepositoryError:
#         pass
# 
# _test_git()



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


def rgb2gray(image):
    """
    Convert an rgb image to grayscale.
    """
    return _np.dot(image, [0.2126, 0.7152, 0.0722])



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np
from .setup_figure import setup_figure as _setup_figure


def setup_axes(rows=1, cols=1, figsize=(8, 6), expand=True, **kwargs):
    """
    .. versionadded:: 1.2

    Sets up a figure of size *figsize* with a number of rows (*rows*) and columns (*cols*). \*\*kwargs passed through to :meth:`matplotlib.figure.Figure.add_subplot`.

    Returns :code:`fig, axes`:
    
    * *fig*: The figure
    * *axes*: An array of all of the axes. (Unless there's only one axis, in which case it returns an object instance :class:`matplotlib.axis.Axis`.)
    """

    if expand:
        figsize = (figsize[0]*cols, figsize[1]*rows)

    fig, gs = _setup_figure(rows=rows, cols=cols, figsize=figsize)

    axes = _np.empty(shape=(rows, cols), dtype=object)

    for i in range(rows):
        for j in range(cols):
            axes[i, j] = fig.add_subplot(gs[i, j], **kwargs)

    if axes.shape == (1, 1):
        return fig, axes[0, 0]
    else:
        return fig, axes



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import matplotlib.pyplot as _plt
    import matplotlib as _mpl
    import numpy as _np
    import pdb as pdb

from .colorbar import colorbar as _cb
from .setup_axes import setup_axes as _setup_axes

_CONTOUR = 1
_IMSHOW  = 2
_QUIVER  = 3

__all__ = [
    'contour',
    'imshow',
    'scaled_figsize'
    ]


def imshow(X, ax=None, add_cbar=True, rescale_fig=True, **kwargs):
    """
    .. versionadded:: 1.3

    Plots an array *X* such that the first coordinate is the *x* coordinate and the second coordinate is the *y* coordinate, with the origin at the bottom left corner.

    Optional argument *ax* allows an existing axes to be used.

    *\*\*kwargs* are passed on to :meth:`matplotlib.axes.Axes.imshow`.

    * Returns :code:`fig, ax, im` if axes aren't specified.
    * Returns :code:`im` if axes are specified.
    """
    return _plot_array(X, plottype=_IMSHOW, ax=ax, add_cbar=add_cbar, rescale_fig=rescale_fig, **kwargs)


def contour(X, ax=None, add_cbar=True, rescale_fig=True, **kwargs):
    """
    .. versionadded:: 1.3

    Plots an array *X* such that the first coordinate is the *x* coordinate and the second coordinate is the *y* coordinate, with the origin at the bottom left corner.

    Optional argument *ax* allows an existing axes to be used.

    *\*\*kwargs* are passed on to :meth:`matplotlib.axes.Axes.contour`.

    Returns :class:`matplotlib.image.AxesImage`.
    """
    return _plot_array(X, plottype=_CONTOUR, ax=ax, add_cbar=add_cbar, rescale_fig=rescale_fig, **kwargs)


def quiver(*args, ax=None, rescale_fig=True, **kwargs):
    """
    .. versionadded:: 1.3

    Plots an array *X* such that the first coordinate is the *x* coordinate and the second coordinate is the *y* coordinate, with the origin at the bottom left corner.

    Optional argument *ax* allows an existing axes to be used.

    *\*\*kwargs* are passed on to :meth:`matplotlib.axes.Axes.quiver`.

    Returns :class:`matplotlib.image.AxesImage`.
    """
    return _plot_array(*args, plottype=_QUIVER, ax=ax, add_cbar=False, rescale_fig=rescale_fig, **kwargs)


def _plot_array(*args, plottype, ax=None, add_cbar=True, rescale_fig=True, **kwargs):
    # ======================================
    # Get an ax
    # ======================================
    if ax is None:
        if rescale_fig:
            figsize = scaled_figsize(args[0])
        else:
            figsize = None
        fig, ax_h = _setup_axes(figsize=figsize)
    else:
        ax_h = ax

    if plottype == _IMSHOW:
        im = ax_h.imshow(_np.transpose(*args), origin='lower', **kwargs)
    elif plottype == _CONTOUR:
        im = ax_h.contour(_np.transpose(*args), origin='lower', **kwargs)
    elif plottype == _QUIVER:
        if len(args) == 2:
            im = ax_h.quiver(_np.transpose(args[0]), _np.transpose(args[1]), **kwargs)
        elif len(args) == 4:
            x_new = _np.transpose(args[0])
            y_new = _np.transpose(args[1])
            u_new = _np.transpose(args[2])
            v_new = _np.transpose(args[3])
            im = ax_h.quiver(x_new, y_new, u_new, v_new, **kwargs)
        else:
            raise NotImplementedError('Only quiver(U, V, **kw) and quiver(X, Y, U, V, *kw) supported at the moment.')

    if add_cbar:
        cb = _cb(ax_h, im)

    if ax is None:
        return fig, ax_h, im
    else:
        if add_cbar:
            return im, cb
        else:
            return im


def scaled_figsize(X, figsize=None, h_pad=None, v_pad=None):
    """
    .. versionadded:: 1.3

    Given an array *X*, determine a good size for the figure to be by shrinking it to fit within *figsize*. If not specified, shrinks to fit the figsize specified by the current :attr:`matplotlib.rcParams`.
    """
    if figsize is None:
        figsize = _mpl.rcParams['figure.figsize']

    # ======================================
    # Find the height and width
    # ======================================
    width, height = _np.shape(X)
    
    ratio = width / height
    
    # ======================================
    # Find how to rescale the figure
    # ======================================
    if ratio > figsize[0]/figsize[1]:
        figsize[1] = figsize[0] / ratio
    else:
        figsize[0] = figsize[1] * ratio
    
    return figsize



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


def less_labels(ax, x_fraction=0.5, y_fraction=0.5):
    """
    Scale the number of tick labels in x and y by *x_fraction* and *y_fraction* respectively.
    """
    nbins = _np.size(ax.get_xticklabels())
    ax.locator_params(nbins=_np.floor(nbins*x_fraction), axis='x')

    nbins = _np.size(ax.get_yticklabels())
    ax.locator_params(nbins=_np.floor(nbins*y_fraction), axis='y')



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np


def piecewise(x, condlist, funclist, *args, **kw):
    x = _np.array([x], dtype=float).flatten()

    return _np.piecewise(x, condlist, funclist, *args, **kw)



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import scipy.constants as _spc
    import scipy.special as _spsp
    import numpy as _np
    import scipy as _sp

    # ============================
    # Constants
    # ============================
    e      = _spc.e               # Elementary charge
    c      = _spc.speed_of_light  # Speed of light
    e0     = _spc.epsilon_0       # Vacuum permittivity
    amu    = _spc.atomic_mass     # AMU in kg

from .ions import Ions as _Ions
import logging as _logging
_logger = _logging.getLogger(__name__)


class Ions1D(_Ions):
    """
    .. versionadded:: 1.7

    A class to facilitate calculating ion motion in PWFA ion columns due to planar, infinitely-long gaussian beams.
    """
    def __init__(self, species, N_e, sig_x, sig_y, sig_xi, r0_big=None, n_samp=1000, order=5, rtol=None, atol=None):
        # ============================
        # Experiment Parameters
        # ============================
        super().__init__(dims=1, species=species, N_e=N_e, sig_xi=sig_xi, rtol=rtol, atol=atol)
        if sig_x < sig_y:
            raise ValueError('Assumption that (sig_x = {}) > (sig_y = {}) not met.'.format(sig_x, sig_y))
        else:
            self._sig_x  = sig_x   # Transverse R.M.S. width
            self._sig_y  = sig_y   # Transverse R.M.S. width

    def lambda_large(self, y0):
        """
        The wavelength for large (:math:`r_0 < \\sigma_r`) oscillations.
        """
        return 4*_np.sqrt(2*y0/self.k)

    # ============================
    # Omega for large r(0)
    # ============================
    # def omega_big(self, r0):
    #     omega_r0 = _np.sqrt(_np.pi*self.k/2)
    #     return omega_r0/_np.sqrt(r0)

    # ============================
    # Scale factor function
    # ============================
    def h(self, q):
        return 1

    @property
    def sig_x(self):
        return self._sig_x

    @property
    def sig_y(self):
        return self._sig_y

    @property
    def nb0(self):
        return self.N_e / (4*_np.sqrt(3) * _np.pi * self.sig_x * self.sig_y * self.sig_xi)

    @property
    def k(self):
        """
        Driving force term: :math:`r'' = -k \\left( \\frac{1-e^{-r^2/2{\\sigma_r}^2}}{r} \\right)`
        """
        try:
            return self._k
        except AttributeError:
            self._k = _np.sqrt(_np.pi/8) * e**2 * self.nb0 * self.sig_y / ( e0 * self.m * c**2)
            return self._k

    @property
    def k_small(self):
        """
        Small-angle driving force term: :math:`r'' = -k_{small} r`.

        Note: :math:`k_{small} = \\frac{k}{2{\\sigma_r^2}}`
        """
        return self.k * _np.sqrt(2/_np.pi) / self.sig_y
        # return e**2 * self.nb0 / (2 * e0 * self.m * c**2)

    # ============================
    # Force equation
    # ============================
    def _yp2(self, y):
        return -self.k*_spsp.erf(y/(_np.sqrt(2)*self.sig_y))
    
    # ============================
    # Derivative of force eq.
    # ============================
    def _delyp2(self, y):
        return - (self.k * _np.sqrt(2/_np.pi)/self.sig_y) * _np.exp(-y**2/(2*self.sig_y**2))

    # ============================
    # First-order ODE system
    # ============================
    def _func(self, y, t):
        return [self._yp2(y[1]), y[0]]

    # ============================
    # Gradient for 1st-order ODE
    # ============================
    def _gradient(self, y, t):
        return [[0, self._delyp2(y[1])], [1, 0]]

    @property
    def q_label(self):
        return 'y'



__all__ = [
    'get_app',
    'Mpl_Image',
    'Mpl_Image_Plus_Slider',
    'Mpl_Plot',
    'Slider_and_Text',
    'Button',
    'getDouble',
    'ButtonMsg',
    'getOpenFileName',
    'getExistingDirectory'
    ]
from .get_app import get_app                                                         # noqa
from .mplwidget import Mpl_Image, Mpl_Image_Plus_Slider, Mpl_Plot, Slider_and_Text   # noqa
from .ButtonMsg import *                                                             # noqa
from .Rectangle import Rectangle                                                     # noqa



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import scipy.optimize as _spopt
from .chisquare import chisquare as _chisquare


def curve_fit_unscaled(*args, **kwargs):
    """
    Use the reduced chi square to unscale :mod:`scipy`'s scaled :func:`scipy.optimize.curve_fit`. *\*args* and *\*\*kwargs* are passed through to :func:`scipy.optimize.curve_fit`. The tuple *popt, pcov, chisq_red* is returned, where *popt* is the optimal values for the parameters, *pcov* is the estimated covariance of *popt*, and *chisq_red* is the reduced chi square. See http://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.optimize.curve_fit.html.
    """
    # Extract verbosity
    verbose = kwargs.pop('verbose', False)

    # Do initial fit
    popt, pcov = _spopt.curve_fit(*args, **kwargs)

    # Expand positional arguments
    func = args[0]
    x    = args[1]
    y    = args[2]

    ddof = len(popt)

    # Try to use sigma to unscale pcov
    try:
        sigma = kwargs['sigma']
        if sigma is None:
            sigma = _np.ones(len(y))
        # Get reduced chi-square
        y_expect = func(x, *popt)
        chisq_red = _chisquare(y, y_expect, sigma, ddof, verbose=verbose)

        # Correct scaled covariance matrix
        pcov = pcov / chisq_red
        return popt, pcov, chisq_red
    except ValueError:
        print('hello')
# popt, pcov = _spopt.curve_fit(func, x, y, sigma=sigma_y, p0=p0)

# # Get reduced chi-square
# y_expect = func(x, popt[0], popt[1], popt[2], popt[3])
# chisq_red = _chisquare(y, y_expect, sigma_y, 3, verbose=verbose)

# # Correct scaled covariance matrix
# pcov = pcov/chisq_red



import sys as _sys
import time as _time
import datetime as _dt


class progressbar(object):
    """
    .. versionadded:: 1.3

    Creates a progress bar with a number of steps *total* and a length of *length*.
    """
    def __init__(self, total, length=20):
        print('')
        self._step   = 0
        self._total  = total
        self._length = length
        self._timestamp = None
        self._timestart = None

    @property
    def step(self):
        """
        The current step.
        """
        return self._step

    @step.setter
    def step(self, step):
        lasttime = self._timestamp
        self._timestamp = _time.perf_counter()
        if lasttime is not None:
            dt = self._timestamp-lasttime
            time_remain = (self._total - step - 1) * dt
            remain_str = ', {} remain'.format(str(_dt.timedelta(seconds=time_remain)))
        else:
            remain_str = ''

        if self._timestart is None:
            self._timestart = self._timestamp
            elapsed_str = ''
        else:
            elapsed = self._timestamp-self._timestart
            elapsed_str = ', {} elapsed'.format(str(_dt.timedelta(seconds=elapsed)))

        if step > self._total:
            step = self._total
        self._step = step
        step = step - 1
        bartext = '#'*round(step/self._total * self._length) + ' '*round((self._total-step)/self._total * self._length)
        text = '\r\033[1AOn step {} of {} ({:0.1f}% completed{}{}):\n[ {} ]'.format(self._step, self._total, 100.0*step/self._total, remain_str, elapsed_str, bartext)
        _sys.stdout.write(text)
        _sys.stdout.flush()
        # print(text)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        text = '\r\033[K\033[1A\033[K'
        _sys.stdout.write(text)
        _sys.stdout.flush()



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    from mpl_toolkits import axes_grid1 as _ag1
    import matplotlib.pyplot as _plt

import logging as _logging
_logger = _logging.getLogger(__name__)

import pdb as _pdb


def colorbar(ax, im, fig=None, loc="right", size="5%", pad="3%"):
    """
    .. versionadded:: 1.3

    Adds a polite colorbar that steals space so :meth:`matplotlib.pyplot.tight_layout` works nicely.
    """
    if fig is None:
        fig = ax.get_figure()

    # _pdb.set_trace()
    if loc == "left" or loc == "right":
        width = fig.get_figwidth()
        new = width * (1 + _pc2f(size) + _pc2f(pad))
        _logger.debug('Setting new figure width: {}'.format(new))
        fig.set_size_inches(new, fig.get_figheight(), forward=True)
    elif loc == "top" or loc == "bottom":
        height = fig.get_figheight()
        new = height * (1 + _pc2f(size) + _pc2f(pad))
        _logger.debug('Setting new figure height: {}'.format(new))
        fig.set_figheight(fig.get_figwidth(), new, forward=True)

    divider = _ag1.make_axes_locatable(ax)
    cax = divider.append_axes(loc, size=size, pad=pad)
    return cax, _plt.colorbar(im, cax=cax)


def _pc2f(str):
    return float(str.strip('%'))/100



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    from matplotlib.pyplot import figure as _figure


def figure(title=None, **kwargs):
    """
    Creates a figure with *\*\*kwargs* with a window title *title*.

    Returns class :class:`matplotlib.figure.Figure`.
    """
    fig = _figure(**kwargs)
    if title is not None:
        fig.canvas.set_window_title(title)
    return fig



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.pyplot as _plt
    from matplotlib import gridspec as _gridspec


def setup_figure(rows=1, cols=1, **kwargs):
    """
    .. versionchanged:: 1.3
       Supports *\*\*kwargs* pass-through to :class:`matplotlib.figure.Figure`.
       
    .. versionchanged:: 1.2
       Changed *gridspec_x* to *rows*, *gridspec_y* to *cols*, added *figsize* control.


    Sets up a figure with a number of rows (*rows*) and columns (*cols*), *\*\*kwargs* passes through to :class:`matplotlib.figure.Figure`.

    Returns :code:`fig, gs`:

    * *fig*: :class:`matplotlib.figure.Figure` instance
    * *gs*: :class:`matplotlib.gridspec.GridSpec` instance with *gridspec_x* rows and *gridspec_y* columns
    """
    fig = _plt.figure(**kwargs)
    gs = _gridspec.GridSpec(rows, cols)

    return fig, gs



import logging
loggerlevel = logging.DEBUG
logger      = logging.getLogger(__name__)

from ..BDES2K import K2BDES, BDES2K
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as np
energy0    = np.float(20.35)
QS1_length = np.float(1)
QS2_length = np.float(1)

__all__ = ['setQS']


class QS(object):
    def __init__(self, K1, length, energy):
        self.K1     = K1
        self.energy = energy
        self.length = length

    def _get_BDES(self):
        return K2BDES(K=self.K1, quad_length=self.length, energy=self.energy)

    def _set_BDES(self, value):
        return BDES2K(BDES=value, quad_length=self.length, energy=self.energy)
    BDES = property(_get_BDES, _set_BDES)
    

class setQS(object):
    def __init__(self, energy_offset):
        logger.critical('****************************USING HARDCODED FUNCTIONS!!!****************************')
        logger.log(level=loggerlevel, msg='Energy offset: {}'.format(energy_offset))
        self.energy_offset = energy_offset
        
    # ======================================
    # QS1, QS2 class objects
    # ======================================
    def _get_QS1(self):
        out = QS(
            K1     = self._get_QS1_K1(),
            length = QS1_length,
            energy = self.energy
            )
        return out
    QS1 = property(_get_QS1)
        
    def _get_QS2(self):
        out = QS(
            K1     = self._get_QS2_K1(),
            length = QS2_length,
            energy = self.energy
            )
        return out
    QS2 = property(_get_QS2)

    # ======================================
    # Get QS1/2's K1 value
    # ======================================
    def _get_QS1_K1(self):
        QS1_BDES = self._get_QS1_BDES()
        return BDES2K(bdes=QS1_BDES, quad_length = QS1_length, energy=self.energy)

    def _get_QS2_K1(self):
        QS2_BDES = self._get_QS2_BDES()
        return BDES2K(bdes=QS2_BDES, quad_length = QS2_length, energy=self.energy)

    # ======================================
    # Get QS1/2's BDES value
    # ======================================
    def _get_QS1_BDES(self):
        QS1_BDES = (np.float64(1)+self.energy_offset/np.float64(20.35))*np.float64(261.72)
        return QS1_BDES

    def _get_QS2_BDES(self):
        QS2_BDES = -(np.float64(1)+self.energy_offset/np.float64(20.35))*np.float64(167.95)
        return QS2_BDES

    # ======================================
    # Get energy in GeV
    # ======================================
    def _get_energy(self):
        return self.energy_offset + energy0

    def _set_energy(self, value):
        self.energy_offset = value - energy0
    energy = property(_get_energy, _set_energy)

    # ======================================
    # Get ELANEX's y motor position
    # ======================================
    def elanex_y_motor(self):
        ymotor = np.float64(55.)*self.energy_offset/(self.energy_offset+np.float64(20.35))
        logger.critical('Ymotor is: {}'.format(ymotor))

        return ymotor



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import matplotlib.pyplot as _plt
    import numpy as _np

from .setup_axes import setup_axes as _setup_axes


def plot(*args, ax=None, **kwargs):
    """
    .. versionadded:: 1.4

    Plots but automatically resizes x axis.
    """
    if ax is None:
        fig, ax = _setup_axes()

    pl = ax.plot(*args, **kwargs)

    if _np.shape(args)[0] > 1:
        if type(args[1]) is not str:
            min_x = min(args[0])
            max_x = max(args[0])
            ax.set_xlim((min_x, max_x))

    return pl



#!/usr/local/lcls/package/python/current/bin/python -d
# encoding: utf-8

'''
Name: logbookForm.py
Author: Shawn Alverson
Created: 2/01/2013

Purpose: Logbook entry form for use with canvas.py
'''

from PyQt4.QtCore import *
from PyQt4.QtGui import *
__all__ = ['logbookForm', 'sendToLogbook']


class logbookForm(QDialog):
    """Dialog GUI for creating logbook entries for both MCC and Physics(DESY) type logbooks"""
    
    def __init__(self, parent=None, image=None, logType='MCC', log='LCLS', deleteOnClose=True):
        from .ui_logbookForm import Ui_logbookForm  # imports Qt gui design file
        
        QDialog.__init__(self, parent)
        if deleteOnClose:
            # If set, object is deleted when close signal is sent
            self.setAttribute(Qt.WA_DeleteOnClose, deleteOnClose)

        self.logui = Ui_logbookForm()
        self.logui.setupUi(self)

        # Initialize variables
        self.mcc_programs = []
        self.physics_programs = []
        self.logMenus = []
        self.logMenuCount = 0
        self.imageType = "png"
        
        # QDir.setCurrent('/u/ad/alverson/cvswork/tools/python/toolbox')
        file = open(self.resourcePath('StyleSheets/LightTheme/lightTheme.qss'), 'r')
        stylesheet = file.read()
        stylesheet.replace('%%', self.resourcePath('StyleSheets/LightTheme/'))
        self.setStyleSheet(stylesheet)
        
        # Check is image is provided and of valid type
        self.image = image
        self.checkImage(image)

        # Set up Logbook pulldowns
        self.logTypeList = ["MCC", "Physics"]
        self.getLogbooks()
        self._connectSlots()
        mccDef = "MCC"
        physDef = "LCLS"
        if logType == "MCC":
            mccDef = log
        else:
            physDef = log
        self.addLogbook(physDef, mccDef, initialInstance=True)

        # Select default log in pulldown
        self.logMenus[0].logType.setCurrentIndex(self.logMenus[0].logType.findText(logType))

    def resourcePath(self, relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        from os import path
        import sys
        
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            base_path = path.dirname(path.abspath(__file__))
        return path.join(base_path, relative_path)
    
    def _connectSlots(self):
        QObject.connect(self.logui.submitButton, SIGNAL("clicked()"), self.submitEntry)
        QObject.connect(self.logui.cancelButton, SIGNAL("clicked()"), self.reject)
        QObject.connect(self.logui.resetButton, SIGNAL("clicked()"), self.clearForm)

    def checkImage(self, image):
        if type(image) != QPixmap:
            if image is None:
                image = QPixmap()
            elif type(image) == str or type(image) == str:
                from os import path
                if path.exists(image):
                    if path.splitext(image)[1] == ".gif":
                        self.imageType = "gif"
                    image = QPixmap(image)
                else:
                    raise Exception("Invalid path to image file specified: " + image)
            else:
                raise Exception("Invalid image type: " + str(type(image)) + "! Must provide QPixmap or path to valid image file.")
        self.imagePixmap = image
        self.pinImage()

    def pinImage(self):
        # Scale image file down to 40x40 icon and paint in thumbnail
        if not self.imagePixmap.isNull():
            icon = self.imagePixmap.scaled(50, 50, Qt.KeepAspectRatio, Qt.FastTransformation)
            self.logui.imageFrame.setPixmap(icon)
        else:
            self.logui.imageFrame.setVisible(False)
    
    def getLogbooks(self):
        from urllib.request import urlopen, URLError, HTTPError
        import json
        
        # Retrieve MCC sublogs,
        while True:
            networkFault = False
            log_url = "https://mccelog.slac.stanford.edu/elog/dev/mgibbs/dev_json_logbook_list.php"
            try:
                data = urlopen(log_url, None, 5).read()
                data = json.loads(data.decode('utf8'))
                break
            except URLError as error:
                print("URLError: " + str(error.reason))
                networkFault = True
            except HTTPError as error:
                print("HTTPError: " + str(error.reason))
                networkFault = True
            
            # Ask to repeat if network connection fails
            if networkFault:
                data = [{"name": "MCC"}, {"name": "SPEAR"}, {"name": "NLCTA"}]
                msgBox = QMessageBox()
                msgBox.setText("Cannot connect to MCC Log Server!")
                msgBox.setInformativeText("Load default logbooks?")
                msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Retry)
                msgBox.setDefaultButton(QMessageBox.Retry)
                if msgBox.exec_() != QMessageBox.Retry:
                    break
        
        self.mcc_programs = []
        for i in data:
            self.mcc_programs.append(str(i["name"]))
        self.mcc_programs.sort()
        
        # Set up Physics logbooks
        self.physics_programs = ["LCLS", "FACET"]
    
    def addLogbook(self, physDef= "LCLS", mccDef="MCC", initialInstance=False):
        '''Add new block of logbook selection windows. Only 5 allowed.'''
        if self.logMenuCount < 5:
            self.logMenus.append(LogSelectMenu(self.logui.multiLogLayout, initialInstance))
            self.logMenus[-1].addLogbooks(self.logTypeList[1], self.physics_programs, physDef)
            self.logMenus[-1].addLogbooks(self.logTypeList[0], self.mcc_programs, mccDef)
            self.logMenus[-1].show()
            self.logMenuCount += 1
            if initialInstance:
                # Initial logbook menu can add additional menus, all others can only remove themselves.
                QObject.connect(self.logMenus[-1].logButton, SIGNAL("clicked()"), self.addLogbook)
            else:
                from functools import partial
                QObject.connect(self.logMenus[-1].logButton, SIGNAL("clicked()"), partial(self.removeLogbook, self.logMenus[-1]))
    
    def removeLogbook(self, menu=None):
        '''Remove logbook menu set.'''
        if self.logMenuCount > 1 and menu is not None:
            menu.removeMenu()
            self.logMenus.remove(menu)
            self.logMenuCount -= 1
    
    def selectedLogs(self):
        '''Return selected log books by type.'''
        mcclogs = []
        physlogs = []
        for i in range(len(self.logMenus)):
            logType = self.logMenus[i].selectedType()
            log = self.logMenus[i].selectedProgram()
            if logType == "MCC":
                if log not in mcclogs:
                    mcclogs.append(log)
            elif logType == "Physics":
                if log not in physlogs:
                    physlogs.append(log)
        return mcclogs, physlogs
    
    def acceptedUser(self, logType):
        '''Verify enetered user name is on accepted MCC logbook list.'''
        from urllib2 import urlopen, URLError, HTTPError
        import json
        
        isApproved = False
        
        userName = str(self.logui.userName.text())
        if userName == "":
            return False  # Must have a user name to submit entry
        
        if logType == "MCC":
            networkFault = False
            data = []
            log_url = "https://mccelog.slac.stanford.edu/elog/dev/mgibbs/dev_json_user_list.php/?username=" + userName
            try:
                data = urlopen(log_url, None, 5).read()
                data = json.loads(data)
            except URLError as error:
                print("URLError: " + str(error.reason))
                networkFault = True
            except HTTPError as error:
                print("HTTPError: " + str(error.reason))
                networkFault = True
            
            # If network fails, ask user to verify
            if networkFault:
                msgBox = QMessageBox()
                msgBox.setText("Cannot connect to MCC Log Server!")
                msgBox.setInformativeText("Use entered User name anyway?")
                msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
                msgBox.setDefaultButton(QMessageBox.Ok)
                if msgBox.exec_() == QMessageBox.Ok:
                    isApproved = True
            
            if data != [] and (data is not None):
                isApproved = True
        else:
            isApproved = True
        return isApproved
    
    def xmlSetup(self, logType, logList):
        """Create xml file with fields from logbook form."""
        
        from xml.etree.ElementTree import Element, SubElement, ElementTree
        from datetime import datetime
        
        curr_time = datetime.now()
        if logType == "MCC":
            # Set up xml tags
            log_entry = Element('log_entry')
            title     = SubElement(log_entry, 'title')
            program   = SubElement(log_entry, 'program')
            timestamp = SubElement(log_entry, 'timestamp')
            priority  = SubElement(log_entry, 'priority')
            os_user   = SubElement(log_entry, 'os_user')
            hostname  = SubElement(log_entry, 'hostname')
            text      = SubElement(log_entry, 'text')
            log_user  = SubElement(log_entry, 'log_user')

            # Check for multiple logbooks and parse into seperate tags
            logbook = []
            for i in range(len(logList)):
                logbook.append(SubElement(log_entry, 'logbook'))
                logbook[i].text = logList[i].lower()
                           
            # Take care of dummy, unchanging tags first
            log_entry.attrib['type'] = "LOGENTRY"
            program.text = "152"
            priority.text = "NORMAL"
            os_user.text = "nobody"
            hostname.text = "mccelog"
            text.attrib['type'] = "text/plain"
            
            # Handle attachment if image exists
            if not self.imagePixmap.isNull():
                attachment = SubElement(log_entry, 'attachment')
                attachment.attrib['name'] = "Figure 1"
                attachment.attrib['type'] = "image/" + self.imageType
                attachment.text = curr_time.strftime("%Y%m%d_%H%M%S_") + str(curr_time.microsecond) + "." + self.imageType
            
            # Set timestamp format
            timestamp.text = curr_time.strftime("%Y/%m/%d %H:%M:%S")
            
            fileName = "/tmp/" + curr_time.strftime("%Y%m%d_%H%M%S_") + str(curr_time.microsecond) + ".xml"
            
        else:  # If using Physics logbook
            timeString = curr_time.strftime("%Y-%m-%dT%H:%M:%S")
            
            # Set up xml tags
            log_entry = Element(None)
            severity  = SubElement(log_entry, 'severity')
            location  = SubElement(log_entry, 'location')
            keywords  = SubElement(log_entry, 'keywords')
            time      = SubElement(log_entry, 'time')
            isodate   = SubElement(log_entry, 'isodate')
            log_user  = SubElement(log_entry, 'author')
            category  = SubElement(log_entry, 'category')
            title     = SubElement(log_entry, 'title')
            metainfo  = SubElement(log_entry, 'metainfo')
            
            # Handle attachment if image exists
            if not self.imagePixmap.isNull():
                imageFile = SubElement(log_entry, 'link')
                imageFile.text = timeString + "-00." + self.imageType
                thumbnail = SubElement(log_entry, 'file')
                thumbnail.text = timeString + "-00.png"
                
            text      = SubElement(log_entry, 'text')  # Logbook expects Text tag to come last (for some strange reason)
            
            # Take care of dummy, unchanging tags first
            log_entry.attrib['type'] = "LOGENTRY"
            category.text = "USERLOG"
            location.text = "not set"
            severity.text = "NONE"
            keywords.text = "none"
            
            time.text = curr_time.strftime("%H:%M:%S")
            isodate.text = curr_time.strftime("%Y-%m-%d")
            
            metainfo.text = timeString + "-00.xml"
            fileName = "/tmp/" + metainfo.text
            
        # Fill in user inputs
        log_user.text = str(self.logui.userName.text())

        title.text = str(self.logui.titleEntry.text())
        if title.text == "":
            QMessageBox().warning(self, "No Title entered", "Please enter a title for the entry...")
            return None
            
        text.text = str(self.logui.textEntry.toPlainText())
        # If text field is truly empty, ElementTree leaves off tag entirely which causes logbook parser to fail
        if text.text == "":
            text.text = " "
        
        # Create xml file
        xmlFile = open(fileName, "w")
        if logType == "MCC":
            ElementTree(log_entry).write(xmlFile)
        else:
            xmlString = self.prettify(log_entry)
            xmlFile.write(xmlString)
        xmlFile.write("\n")  # Close with newline so cron job parses correctly
        xmlFile.close()
            
        return fileName.rstrip(".xml")

    def prettify(self, elem):
        """Parse xml elements for pretty printing"""
        
        from xml.etree import ElementTree
        from re import sub
        
        rawString = ElementTree.tostring(elem, 'utf-8')
        parsedString = sub(r'(?=<[^/].*>)', '\n', rawString)  # Adds newline after each closing tag
        
        return parsedString[1:]
    
    def prepareImages(self, fileName, logType):
        """Convert supplied QPixmap object to image file."""
        import subprocess
        
        if self.imageType == "png":
            self.imagePixmap.save(fileName + ".png", "PNG", -1)
            if logType == "Physics":
                makePostScript = "convert " + fileName + ".png " + fileName + ".ps"
                process = subprocess.Popen(makePostScript, shell=True)
                process.wait()
                thumbnailPixmap = self.imagePixmap.scaled(500, 450, Qt.KeepAspectRatio)
                thumbnailPixmap.save(fileName + ".png", "PNG", -1)
        else:
            renameImage = "cp " + self.image + " " + fileName + ".gif"
            process = subprocess.Popen(renameImage, shell=True)
            process.wait()
            if logType == "Physics":
                thumbnailPixmap = self.imagePixmap.scaled(500, 450, Qt.KeepAspectRatio)
                thumbnailPixmap.save(fileName + ".png", "PNG", -1)
    
    def submitEntry(self):
        """Process user inputs and subit logbook entry when user clicks Submit button"""
        
        # logType = self.logui.logType.currentText()
        mcclogs, physlogs = self.selectedLogs()
        success = True
        
        if mcclogs != []:
            if not self.acceptedUser("MCC"):
                QMessageBox().warning(self, "Invalid User", "Please enter a valid user name!")
                return
            
            fileName = self.xmlSetup("MCC", mcclogs)
            if fileName is None:
                return
            
            if not self.imagePixmap.isNull():
                self.prepareImages(fileName, "MCC")
            success = self.sendToLogbook(fileName, "MCC")
        
        if physlogs != []:
            for i in range(len(physlogs)):
                fileName = self.xmlSetup("Physics", physlogs[i])
                if fileName is None:
                    return
            
                if not self.imagePixmap.isNull():
                    self.prepareImages(fileName, "Physics")
                success_phys = self.sendToLogbook(fileName, "Physics", physlogs[i])
                success = success and success_phys
            
        self.done(success)
    
    def sendToLogbook(self, fileName, logType, location=None):
        '''Process log information and push to selected logbooks.'''
        import subprocess
        
        success = True
        if logType == "MCC":
            fileString = ""
            if not self.imagePixmap.isNull():
                fileString = fileName + "." + self.imageType
        
            logcmd = "xml2elog " + fileName + ".xml " + fileString
            process = subprocess.Popen(logcmd, shell=True)
            process.wait()
            if process.returncode != 0:
                success = False
        else:
            from shutil import copy

            path = "/u1/" + location.lower() + "/physics/logbook/data/"  # Prod path
            # path = "/home/softegr/alverson/log_test/"  # Dev path
            try:
                if not self.imagePixmap.isNull():
                    copy(fileName + ".png", path)
                    if self.imageType == "png":
                        copy(fileName + ".ps", path)
                    else:
                        copy(fileName + "." + self.imageType, path)
            
                # Copy .xml file last to ensure images will be picked up by cron job
                # print("Copying file " + fileName + " to path " + path)
                copy(fileName + ".xml", path)
            except IOError as error:
                print(error)
                success = False
            
        return success
    
    def clearForm(self):
        """Clear all form fields (except author)."""
        
        self.logui.titleEntry.clear()
        self.logui.textEntry.clear()

        # Remove all log selection menus except the first
        while self.logMenuCount > 1:
            self.removeLogbook(self.logMenus[-1])


class LogSelectMenu(QWidget):
    '''Menu objects in horizontal layout for selecting logbook facilities'''
    selectedLogs = []
    menuCount = 0
    
    def __init__(self, parent=QVBoxLayout(), initialInstance=True):
        super(LogSelectMenu, self).__init__()
        
        self.parent = parent
        self.logList = {}
        self.initialInstance = initialInstance
        
        self.setupUI()
    
    def setupUI(self):
        '''Create graphical objects for menus.'''
        
        labelSizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        labelSizePolicy.setHorizontalStretch(0)
        labelSizePolicy.setVerticalStretch(0)
        menuSizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        menuSizePolicy.setHorizontalStretch(0)
        menuSizePolicy.setVerticalStretch(0)
        
        logTypeLayout = QHBoxLayout()
        logTypeLayout.setSpacing(0)
        
        typeLabel = QLabel("Log Type:")
        typeLabel.setMinimumSize(QSize(65, 0))
        typeLabel.setMaximumSize(QSize(65, 16777215))
        typeLabel.setSizePolicy(labelSizePolicy)
        logTypeLayout.addWidget(typeLabel)
        self.logType = QComboBox(self)
        self.logType.setMinimumSize(QSize(100, 0))
        self.logType.setMaximumSize(QSize(150, 16777215))
        menuSizePolicy.setHeightForWidth(self.logType.sizePolicy().hasHeightForWidth())
        self.logType.setSizePolicy(menuSizePolicy)
        logTypeLayout.addWidget(self.logType)
        logTypeLayout.setStretch(1, 6)
        
        programLayout = QHBoxLayout()
        programLayout.setSpacing(0)
        
        programLabel = QLabel("Program:")
        programLabel.setMinimumSize(QSize(60, 0))
        programLabel.setMaximumSize(QSize(60, 16777215))
        programLabel.setSizePolicy(labelSizePolicy)
        programLayout.addWidget(programLabel)
        self.programName = QComboBox(self)
        self.programName.setMinimumSize(QSize(100, 0))
        self.programName.setMaximumSize(QSize(150, 16777215))
        menuSizePolicy.setHeightForWidth(self.programName.sizePolicy().hasHeightForWidth())
        self.programName.setSizePolicy(menuSizePolicy)
        programLayout.addWidget(self.programName)
        programLayout.setStretch(1, 6)
        
        # Initial instance allows adding additional menus, all following menus can only remove themselves.
        if self.initialInstance:
            self.logButton = QPushButton("+", self)
            self.logButton.setToolTip("Add logbook")
        else:
            self.logButton = QPushButton("-")
            self.logButton.setToolTip("Remove logbook")
        
        self.logButton.setMinimumSize(QSize(16, 16))  # 24x24
        self.logButton.setMaximumSize(QSize(16, 16))  # 24x24
        self.logButton.setObjectName("roundButton")
        # self.logButton.setAutoFillBackground(True)
        # region = QRegion(QRect(self.logButton.x()+15, self.logButton.y()+14, 20, 20), QRegion.Ellipse)
        # self.logButton.setMask(region)
        
        self.logButton.setStyleSheet("QPushButton {border-radius: 8px;}")
        
        self._logSelectLayout = QHBoxLayout()
        self._logSelectLayout.setSpacing(6)
        self._logSelectLayout.addLayout(logTypeLayout)
        self._logSelectLayout.addLayout(programLayout)
        self._logSelectLayout.addWidget(self.logButton)
        self._logSelectLayout.setStretch(0, 6)
        self._logSelectLayout.setStretch(1, 6)
    
    def _connectSlots(self):
        '''Connect menu change signals.'''
        # QObject.connect(self.logType, SIGNAL("activated(int)"), self.changeLogType)
        QObject.connect(self.logType, SIGNAL("currentIndexChanged(int)"), self.changeLogType)
    
    def show(self):
        '''Display menus and connect even signals.'''
        self.parent.addLayout(self._logSelectLayout)
        self.menuCount += 1
        self._connectSlots()
    
    def selectedType(self):
        return str(self.logType.currentText())
    
    def selectedProgram(self):
        return str(self.programName.currentText())
    
    def addLogbooks(self, type=None, logs=[], default=""):
        '''Add or change list of logbooks.'''
        if type is not None and len(logs) != 0:
            if type in self.logList:
                for logbook in logs:
                    if logbook not in self.logList.get(type)[0]:
                        # print("Adding log " + " to " + type + " log type.")
                        self.logList.get(type)[0].append(logbook)
            else:
                # print("Adding log type: " + type)
                self.logList[type] = []
                self.logList[type].append(logs)
            
            # If default given, auto-select upon menu creation
            if len(self.logList[type]) > 1 and default != "":
                self.logList.get(type)[1] == default
            else:
                self.logList.get(type).append(default)
            
            self.logType.clear()
            self.logType.addItems(list(self.logList.keys()))
            self.changeLogType()
    
    def removeLogbooks(self, type=None, logs=[]):
        '''Remove unwanted logbooks from list.'''
        if type is not None and type in self.logList:
            if len(logs) == 0 or logs == "All":
                del self.logList[type]
            else:
                for logbook in logs:
                    if logbook in self.logList[type]:
                        self.logList[type].remove(logbook)
            
            self.changeLogType()
    
    def changeLogType(self):
        '''Populate log program list to correspond with log type selection.'''
        logType = self.selectedType()
        programs = self.logList.get(logType)[0]
        default = self.logList.get(logType)[1]
        if logType in self.logList:
            self.programName.clear()
            self.programName.addItems(programs)
            self.programName.setCurrentIndex(programs.index(default))
    
    def addMenu(self):
        '''Add menus to parent gui.'''
        self.parent.multiLogLayout.addLayout(self.logSelectLayout)
        self.getPrograms(logType, programName)
    
    def removeLayout(self, layout):
        '''Iteratively remove graphical objects from layout.'''
        for cnt in reversed(range(layout.count())):
            item = layout.takeAt(cnt)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            else:
                '''If sublayout encountered, iterate recursively.'''
                self.removeLayout(item.layout())
    
    def removeMenu(self):
        '''Remove layout from parent gui and destroy object on completion.'''
        self.removeLayout(self._logSelectLayout)
        self.deleteLater()
    

if __name__ == "__main__":
    '''If not spawned by external GUI, launch from localized app'''
    from sys import argv, exit

    app = QApplication(argv)
    
    image = QPixmap()
    deleteOnClose = True
    argnum = len(argv)
    if argnum > 1:
        image = argv[1]
    if argnum >= 2:
        deleteOnClose = argv[2]
    
    log = logbookForm(None, image, "MCC", "MCC", deleteOnClose)
    log.show()

    exit(app.exec_())



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import matplotlib.pyplot as _plt

from .curve_fit_unscaled import curve_fit_unscaled as _curve_fit_unscaled
from ..matplotlib.figure import figure as _figure
import logging
logger = logging.getLogger(__name__)


class GaussResults(object):
    """
    Fits a gaussian to a curve specified by pairs *x* and *y*, with error on *y* of *sigma_y*.

    * *plot*: Determines whether the plot is shown
    * *p0*: Initial guess given by amplitude *amp*, mean *mu*, and standard deviation *rms*, in the form of:

      * :code:`[amp, mu, rms**2]` if *variance_bool* is true
      * :code:`[amp, mu, rms]` if *variance_bool* is false

    * *background_bool*: If true, uses a background term in the fit, with initial guess *bg*

    Returns full statistical results in the form of an instance of class :class:`GaussResults <scisalt.scipy.GaussResults>`.
    """
    def __init__(self, x, y, sigma_y=None, p0=None, variance=False, background=False):
        # ======================================
        # Store things
        # ======================================
        self._x          = x.flatten()
        self._y          = y.flatten()
        self._sigma_y    = sigma_y
        self._p0         = p0
        self._variance   = variance
        self._background = background
    
        # ======================================
        # Perform curve fit
        # ======================================
        self._popt, self._pcov, self._chisq_red = _curve_fit_unscaled(self.func, self.x, self.y, sigma=self.sigma_y, p0=self.p0)
    
        # ======================================
        # Keep sigma positive
        # ======================================
        self._popt[2] = _np.abs(self._popt[2])
    
    @property
    def popt(self):
        """
        The fit parameters for the fit function.
        """
        return self._popt

    @property
    def pcov(self):
        """
        The covariance for the fit parameters.
        """
        return self._pcov

    @property
    def chisq_red(self):
        """
        The reduced chi square.
        """
        return self._chisq_red

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def sigma_y(self):
        return self._sigma_y

    @property
    def variance(self):
        return self._variance

    @property
    def background(self):
        return self._background

    @property
    def func(self):
        # ======================================
        # Determine whether to use the variance or std dev form
        # in the gaussian equation
        # ======================================
        if self.variance:
            if self.background:
                return _gaussvar
            else:
                return _gaussvar_nobg
        else:
            if self.background:
                return _gauss
            else:
                return _gauss_nobg

    @property
    def use_error(self):
        return ( self.sigma_y is not None )

    @property
    def p0(self):
        # ======================================
        # Determine initial guesses
        # ======================================
        if (self._p0 is None):
            amp = max(self.y)
            mu  = sum(self.x * self.y) / sum(self.y)
            rms = _np.sqrt(sum((self.x-mu)**2 * self.y) / sum(self.y))

            # ======================================
            # Modify initial guess for variance
            # ======================================
            if self.variance:
                self._p0 = _np.array((amp, mu, rms**2))
            else:
                self._p0 = _np.array((amp, mu, rms))

            # ======================================
            # Modify initial guess for background
            # ======================================
            if self.background:
                self._p0 = _np.append(self._p0, min(self.y))

        return self._p0

    def plot(self):
        xmin = min(self.x)
        xmax = max(self.x)
        x_fit = _np.linspace(xmin, xmax, 1000)
        y_fit = self.func(x_fit, *self.popt)
        _figure('MYTOOLS: Gauss Fit Routine')
        if self.use_error:
            sigma_y = self.sigma_y.flatten()
            _plt.errorbar(self.x, self.y, yerr=sigma_y, fmt='o-')
            _plt.plot(x_fit, y_fit)
        else:
            _plt.plot(self.x, self.y, 'o-', x_fit, y_fit)


def _gauss(x, amp, mu, sigma, bg=0):
    # print 'Sigma is {}.'.format(sigma)
    return _np.abs(amp) * _np.exp( -(x - mu)**2 / (2 * sigma**2)) + bg
    # return _np.abs(amp) * _np.exp(-(x - mu)**2 / (2 * sigma**2))


def _gauss_nobg(x, amp, mu, sigma):
    return _gauss(x, amp, mu, sigma)


def _gaussvar(x, amp, mu, variance, bg=0):
    return _np.abs(amp) * _np.exp(-(x - mu)**2 / (2 * variance)) + bg
    # return _np.abs(amp) * _np.exp(-(x - mu)**2 / (2 * variance))


def _gaussvar_nobg(x, amp, mu, variance):
    return _gaussvar(x, amp, mu, variance)



import sys
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    from PyQt4 import QtGui


def get_app(argv=sys.argv):
    global app
    app = QtGui.QApplication.instance()
    if not app:
        app = QtGui.QApplication(argv)
    return app



import logging
logger = logging.getLogger(__name__)

import os
import sys
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.pyplot as plt


def savefig(filename, path="figs", fig=None, ext='eps', verbose=False, **kwargs):
    """
    Save the figure *fig* (optional, if not specified, latest figure in focus) to *filename* in the path *path* with extension *ext*.

    *\*\*kwargs* is passed to :meth:`matplotlib.figure.Figure.savefig`.
    """
    filename       = os.path.join(path, filename)
    final_filename = '{}.{}'.format(filename, ext).replace(" ", "").replace("\n", "")
    final_filename = os.path.abspath(final_filename)

    final_path = os.path.dirname(final_filename)
    if not os.path.exists(final_path):
        os.makedirs(final_path)

    if verbose:
        print('Saving file: {}'.format(final_filename))

    if fig is not None:
        fig.savefig(final_filename, bbox_inches='tight', **kwargs)
    else:
        plt.savefig(final_filename, bbox_inches='tight', **kwargs)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    from PyQt4 import QtGui, QtCore
    import numpy as _np
from .get_app import get_app

__all__ = ['Button', 'getDouble', 'ButtonMsg', 'getOpenFileName', 'getExistingDirectory']


class Button(object):
    def __init__(self, *args, **kwargs):
        self.standardbutton = None
        self.text           = None
        self.ButtonRole     = None

        self.button_default = kwargs.pop('default', False)
        self.button_escape = kwargs.pop('escape', False)

        if kwargs == {}:
            argnum = _np.size(args)
        else:
            argnum  = _np.size(args) + _np.size(kwargs.values())
        allargs = _np.array(args, dtype=object)
        allargs = _np.append(allargs, kwargs.values())

        if argnum == 1:
            if isinstance(allargs[0], QtGui.QMessageBox.StandardButton):
                self.standardbutton = allargs[0]
            else:
                self.text       = allargs[0]
                self.ButtonRole = QtGui.QMessageBox.AcceptRole
        elif argnum == 2:
            self.text       = allargs[0]
            self.ButtonRole = allargs[1]
        else:
            raise ValueError('Incorrect arguments: {}'.format(allargs))


class getDouble(QtGui.QWidget):
    def __init__(self, **kwargs):
        app = get_app()  # noqa
        super(getDouble, self).__init__()

        title = kwargs.pop('title', '')
        text = kwargs.pop('text', 'Double number:')

        self._input, self._ok = QtGui.QInputDialog.getDouble(self, title, text, **kwargs)

    @property
    def input(self):
        if self._ok:
            return self._input
        else:
            raise IOError('User may have cancelled')


class ButtonMsg(QtGui.QMessageBox):
    def __init__(self, maintext, buttons, title=None, infotext=None):
        self._clicked = None
        self._buttons = buttons
        app = get_app()  # noqa

        super(ButtonMsg, self).__init__()

        self.msgbox = QtGui.QMessageBox()
        self.msgbox.setText(maintext)

        if infotext is not None:
            self.msgbox.setInformativeText(infotext)

        if (title is not None):
            self.msgbox.setWindowTitle(title)

        if (buttons is not None):
            self.btnarray = _np.empty(_np.size(buttons), dtype=object)
            for i, val in enumerate(buttons):
                if not isinstance(val, Button):
                    val = Button(val)
                self.btnarray[i] = self._addbutton(val)
                
                if val.button_default:
                    self.msgbox.setDefaultButton(self.btnarray[i])
                if val.button_escape:
                    self.msgbox.setEscapeButton(self.btnarray[i])
        self.msgbox.exec_()

    def _addbutton(self, button):
        if (button.standardbutton is None):
            added_button = self.msgbox.addButton(button.text, button.ButtonRole)
        else:
            added_button = self.msgbox.addButton(button.standardbutton)
        
        return added_button

    def clickedButton(self):
        return self.msgbox.clickedButton()

    @property
    def clickeditem(self):
        return self._buttons[self.clicked]

    @property
    def clicked(self):
        if self._clicked is None:
            for i, val in enumerate(self.btnarray):
                if (self.clickedButton() == val):
                    self._clicked = i
        
        return self._clicked

    @property
    def clickedArray(self):
        clicked = _np.empty(_np.size(self.btnarray), dtype=_np.bool_)
        for i, val in enumerate(self.btnarray):
            clicked[i] = (self.clickedButton() == val)
            
        return clicked


def getOpenFileName(*args, **kwargs):
    app = get_app()  # noqa

    filepath = QtGui.QFileDialog.getOpenFileName(*args, **kwargs)
    if filepath == '':
        errorstr = 'No valid file selected'
        QtGui.QMessageBox.warning(None, errorstr, errorstr)
        raise IOError(errorstr)

    return str(filepath)


def getExistingDirectory(*args, **kwargs):
    app = get_app()  # noqa

    filepath = QtGui.QFileDialog.getExistingDirectory(*args, **kwargs)
    if filepath == '':
        errorstr = 'No valid directory selected'
        QtGui.QMessageBox.warning(None, errorstr, errorstr)
        raise IOError(errorstr)

    return str(filepath)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import matplotlib.pyplot as _plt

from .addlabel import addlabel as _addlabel
from .figure import figure as _figure
from .colorbar import colorbar as _colorbar


def hist2d(x, y, bins=10, labels=None, aspect="auto", plot=True, fig=None, ax=None, interpolation='none', **kwargs):
    """
    Creates a 2-D histogram of data *x*, *y* with *bins*, *labels* = :code:`[title, xlabel, ylabel]`, aspect ration *aspect*. Attempts to use axis *ax* first, then the current axis of *fig*, then the last axis, to use an already-created window.
    
    Plotting (*plot*) is on by default, setting false doesn't attempt to create a figure.

    *interpolation* sets the interpolation type of :meth:`matplotlib.axis.imshow`.

    Returns a handle and extent as :code:`h, extent`
    """
    h_range   = kwargs.pop('range')
    h_normed  = kwargs.pop('normed')
    h_weights = kwargs.pop('normed')

    h, xe, ye = _np.histogram2d(x, y, bins=bins, range=h_range, normed=h_normed, weights=h_weights)
    extent    = [xe[0], xe[-1], ye[0], ye[-1]]
    # fig     = plt.figure()
    if plot:
        if ax is None:
            if fig is None:
                fig = _figure('hist2d')
            ax = fig.gca()
            ax.clear()

        img = ax.imshow(h.transpose(), extent=extent, interpolation=interpolation, aspect=aspect, **kwargs)
        _colorbar(ax=ax, im=img)

        if labels is not None:
            _addlabel(labels[0], labels[1], labels[2])

        # _showfig(fig, aspect)
    return h, extent



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np
    import slactrac as _sltr
    import scipy.constants as _spc


class Match(object):
    """
    Given a *plasma* of type :class:`Plasma <scisalt.PWFA.Plasma>` and a beam of energy *E* in GeV and normalized emittance *emit_n* in SI units, calculates match parameters
    """
    def __init__(self, plasma, E, emit_n):
        self.plasma = plasma
        self.E = E
        self.emit_n = emit_n

    @property
    def gamma(self):
        """
        Relativistic :math:`\\gamma` of beam
        """
        return _sltr.GeV2gamma(self.E)

    @gamma.setter
    def gamma(self, value):
        self.E = _sltr.gamma2GeV(value)

    @property
    def emit_n(self):
        """
        Emittance of beam
        """
        return self.emit * self.gamma

    @emit_n.setter
    def emit_n(self, value):
        self.emit = value / self.gamma

    @property
    def sigma(self):
        """
        Spot size of matched beam :math:`\\left( \\frac{2 E \\varepsilon_0 }{ n_p e^2 } \\right)^{1/4} \\sqrt{\\epsilon}`
        """
        return _np.power(2*_sltr.GeV2joule(self.E)*_spc.epsilon_0 / (self.plasma.n_p * _np.power(_spc.elementary_charge, 2)) , 0.25) * _np.sqrt(self.emit)

    def beta(self, E):
        """
        :math:`\\beta` function of matched beam
        """
        return 1.0 / _np.sqrt(self.plasma.k_ion(E))
        # return 1.0 / _np.sqrt(2)


class MatchPlasma(object):
    """
    Given a beam of energy *E* in GeV with normalized emittance *emit_n* in SI units and spot size *sigma*, calculates plasma parameters
    """
    def __init__(self, E, emit_n, sigma):
        self.E = E
        self.emit_n = emit_n
        self.sigma = sigma

    @property
    def gamma(self):
        """
        Relativistic :math:`\\gamma` of beam
        """
        return _sltr.GeV2gamma(self.E)

    @gamma.setter
    def gamma(self, value):
        self.E = _sltr.gamma2GeV(value)

    @property
    def emit_n(self):
        """
        Emittance of beam
        """
        return self.emit * self.gamma

    @emit_n.setter
    def emit_n(self, value):
        self.emit = value / self.gamma

    @property
    def beta(self):
        return self.sigma**2/self.emit

    @property
    def n_p(self):
        return 2*_sltr.GeV2joule(self.E)*_spc.epsilon_0 / (self.beta*_spc.elementary_charge)**2

    @property
    def n_p_cgs(self):
        return self.n_p*1e-6



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_logbookForm.ui'
#
# Created: Wed Jun 25 14:54:33 2014
#      by: PyQt4 UI code generator 4.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_logbookForm(object):
    def setupUi(self, logbookForm):
        logbookForm.setObjectName(_fromUtf8("logbookForm"))
        logbookForm.resize(549, 189)
        self.gridLayout = QtGui.QGridLayout(logbookForm)
        self.gridLayout.setContentsMargins(4, 4, 4, 0)
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        self.headerLayout = QtGui.QHBoxLayout()
        self.headerLayout.setObjectName(_fromUtf8("headerLayout"))
        self.userLayout = QtGui.QHBoxLayout()
        self.userLayout.setSpacing(0)
        self.userLayout.setSizeConstraint(QtGui.QLayout.SetDefaultConstraint)
        self.userLayout.setContentsMargins(0, -1, 0, -1)
        self.userLayout.setObjectName(_fromUtf8("userLayout"))
        self.userLabel = QtGui.QLabel(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.userLabel.sizePolicy().hasHeightForWidth())
        self.userLabel.setSizePolicy(sizePolicy)
        self.userLabel.setMinimumSize(QtCore.QSize(40, 0))
        self.userLabel.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.userLabel.setObjectName(_fromUtf8("userLabel"))
        self.userLayout.addWidget(self.userLabel)
        self.userName = QtGui.QLineEdit(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(6)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.userName.sizePolicy().hasHeightForWidth())
        self.userName.setSizePolicy(sizePolicy)
        self.userName.setMinimumSize(QtCore.QSize(120, 0))
        self.userName.setMaximumSize(QtCore.QSize(500, 16777215))
        self.userName.setSizeIncrement(QtCore.QSize(2, 0))
        self.userName.setObjectName(_fromUtf8("userName"))
        self.userLayout.addWidget(self.userName)
        self.userLayout.setStretch(0, 0)
        self.userLayout.setStretch(1, 6)
        self.headerLayout.addLayout(self.userLayout)
        spacerItem = QtGui.QSpacerItem(0, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.headerLayout.addItem(spacerItem)
        self.multiLogLayout = QtGui.QVBoxLayout()
        self.multiLogLayout.setObjectName(_fromUtf8("multiLogLayout"))
        self.headerLayout.addLayout(self.multiLogLayout)
        self.headerLayout.setStretch(0, 6)
        self.headerLayout.setStretch(1, 6)
        self.headerLayout.setStretch(2, 6)
        self.gridLayout.addLayout(self.headerLayout, 0, 0, 1, 1)
        self.titleLayout = QtGui.QHBoxLayout()
        self.titleLayout.setSpacing(0)
        self.titleLayout.setObjectName(_fromUtf8("titleLayout"))
        self.titleLabel = QtGui.QLabel(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.titleLabel.sizePolicy().hasHeightForWidth())
        self.titleLabel.setSizePolicy(sizePolicy)
        self.titleLabel.setMinimumSize(QtCore.QSize(40, 0))
        self.titleLabel.setObjectName(_fromUtf8("titleLabel"))
        self.titleLayout.addWidget(self.titleLabel)
        self.titleEntry = QtGui.QLineEdit(logbookForm)
        self.titleEntry.setObjectName(_fromUtf8("titleEntry"))
        self.titleLayout.addWidget(self.titleEntry)
        self.gridLayout.addLayout(self.titleLayout, 1, 0, 1, 1)
        self.textLayout = QtGui.QHBoxLayout()
        self.textLayout.setSpacing(0)
        self.textLayout.setObjectName(_fromUtf8("textLayout"))
        self.textLabel = QtGui.QLabel(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.textLabel.sizePolicy().hasHeightForWidth())
        self.textLabel.setSizePolicy(sizePolicy)
        self.textLabel.setMinimumSize(QtCore.QSize(40, 0))
        self.textLabel.setObjectName(_fromUtf8("textLabel"))
        self.textLayout.addWidget(self.textLabel)
        self.textEntry = QtGui.QTextEdit(logbookForm)
        self.textEntry.setObjectName(_fromUtf8("textEntry"))
        self.textLayout.addWidget(self.textEntry)
        self.gridLayout.addLayout(self.textLayout, 2, 0, 1, 1)
        self.footerLayout = QtGui.QHBoxLayout()
        self.footerLayout.setSpacing(10)
        self.footerLayout.setContentsMargins(-1, -1, -1, 10)
        self.footerLayout.setObjectName(_fromUtf8("footerLayout"))
        spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Minimum)
        self.footerLayout.addItem(spacerItem1)
        self.imageFrame = QtGui.QLabel(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.imageFrame.sizePolicy().hasHeightForWidth())
        self.imageFrame.setSizePolicy(sizePolicy)
        self.imageFrame.setMinimumSize(QtCore.QSize(50, 50))
        self.imageFrame.setMaximumSize(QtCore.QSize(50, 50))
        self.imageFrame.setFrameShadow(QtGui.QFrame.Plain)
        self.imageFrame.setLineWidth(0)
        self.imageFrame.setText(_fromUtf8(""))
        self.imageFrame.setObjectName(_fromUtf8("imageFrame"))
        self.footerLayout.addWidget(self.imageFrame)
        spacerItem2 = QtGui.QSpacerItem(93, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.footerLayout.addItem(spacerItem2)
        self.resetButton = QtGui.QPushButton(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.resetButton.sizePolicy().hasHeightForWidth())
        self.resetButton.setSizePolicy(sizePolicy)
        self.resetButton.setMinimumSize(QtCore.QSize(80, 0))
        self.resetButton.setMaximumSize(QtCore.QSize(80, 16777215))
        self.resetButton.setObjectName(_fromUtf8("resetButton"))
        self.footerLayout.addWidget(self.resetButton)
        spacerItem3 = QtGui.QSpacerItem(30, 20, QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Minimum)
        self.footerLayout.addItem(spacerItem3)
        self.cancelButton = QtGui.QPushButton(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cancelButton.sizePolicy().hasHeightForWidth())
        self.cancelButton.setSizePolicy(sizePolicy)
        self.cancelButton.setMinimumSize(QtCore.QSize(80, 0))
        self.cancelButton.setMaximumSize(QtCore.QSize(80, 16777215))
        self.cancelButton.setObjectName(_fromUtf8("cancelButton"))
        self.footerLayout.addWidget(self.cancelButton)
        self.submitButton = QtGui.QPushButton(logbookForm)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.submitButton.sizePolicy().hasHeightForWidth())
        self.submitButton.setSizePolicy(sizePolicy)
        self.submitButton.setMinimumSize(QtCore.QSize(80, 0))
        self.submitButton.setMaximumSize(QtCore.QSize(80, 16777215))
        self.submitButton.setDefault(True)
        self.submitButton.setObjectName(_fromUtf8("submitButton"))
        self.footerLayout.addWidget(self.submitButton)
        self.gridLayout.addLayout(self.footerLayout, 3, 0, 1, 1)

        self.retranslateUi(logbookForm)
        QtCore.QMetaObject.connectSlotsByName(logbookForm)

    def retranslateUi(self, logbookForm):
        logbookForm.setWindowTitle(_translate("logbookForm", "Logbook Entry", None))
        self.userLabel.setText(_translate("logbookForm", "User:", None))
        self.titleLabel.setText(_translate("logbookForm", "Title:", None))
        self.textLabel.setText(_translate("logbookForm", "Text:", None))
        self.resetButton.setText(_translate("logbookForm", "Reset", None))
        self.cancelButton.setText(_translate("logbookForm", "Cancel", None))
        self.submitButton.setText(_translate("logbookForm", "Submit", None))




#!/usr/bin/env python3

import os as _os
import argparse as _argparse
import shlex
from subprocess import call

__all__ = ['githubtunnel']


def githubtunnel(user1, server1, user2, server2, port, verbose, stanford=False):
    """
    Opens a nested tunnel, first to *user1*@*server1*, then to *user2*@*server2*, for accessing on *port*.

    If *verbose* is true, prints various ssh commands.

    If *stanford* is true, shifts ports up by 1.

    Attempts to get *user1*, *user2* from environment variable ``USER_NAME`` if called from the command line.
    """
    if stanford:
        port_shift = 1
    else:
        port_shift = 0

    # command1 = 'ssh -nNf -L {}:quickpicmac3.slac.stanford.edu:22 {}@{}'.format(port, user, server)
    command1 = 'ssh -nNf -L {}:{}:22 {}@{}'.format(port-1-port_shift, server2, user1, server1)
    command2 = 'ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -nNf -L {}:cardinal.stanford.edu:22 -p {} {}@localhost'.format(port-port_shift, port-port_shift-1, user2)
    command3 = 'ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -nNf -L {}:github.com:22 -p {} {}@localhost'.format(port, port-1, user2)
    if verbose:
        print(command1)
        if stanford:
            print(command2)
        print(command3)
    try:
        call(shlex.split(command1))
        if stanford:
            call(shlex.split(command2))
        call(shlex.split(command3))
    except:
        print('Failure!')
        pass


# ================================
# Get info one way or another
# ================================
def environvar(prompt, val):
    if val is None:
        val = input(prompt + ' ')

    return val


def _script():
    # ================================
    # Access command line arguments
    # ================================
    parser = _argparse.ArgumentParser(description=
            'Creates a tunnel primarily for Git.')
    parser.add_argument('-V', action='version', version='%(prog)s v0.1')
    parser.add_argument('-v', '--verbose', action='store_true',
            help='Verbose mode.')
    parser.add_argument('-p', '--port', default=7777, type=int,
            help='Local port to listen on.')
    parser.add_argument('-s1', '--server1', default='mcclogin',
            help='First server hop.')
    parser.add_argument('-s2', '--server2', default='iris.slac.stanford.edu',
            help='Second server hop.')
    parser.add_argument('-su', '--stanford', action='store_true',
            help='Tunnel through Stanford')
    parser.add_argument('-u1', '--user1', default=_os.environ.get('PHYSICS_USER'),
            help='User name to use for login.')
    parser.add_argument('-u2', '--user2', default=_os.environ.get('PHYSICS_USER'),
            help='User name to use for login.')

    arg = parser.parse_args()

    # ================================
    # Ask user for logins if needed
    # ================================
    prompt1 = 'User name for server {}?'.format(arg.server1)
    prompt2 = 'User name for server {}?'.format(arg.server2)
    user1 = environvar(prompt1, arg.user1)
    user2 = environvar(prompt2, arg.user2)

    # ================================
    # Run with command line arguments
    # ================================
    githubtunnel(user1, arg.server1, user2, arg.server2, arg.port, arg.verbose, stanford=arg.stanford)

if __name__ == '__main__':
    _script()



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import matplotlib.pyplot as _plt
import logging as _logging
logger = _logging.getLogger(__name__)


def addlabel(ax=None, toplabel=None, xlabel=None, ylabel=None, zlabel=None, clabel=None, cb=None, windowlabel=None, fig=None, axes=None):
    """Adds labels to a plot."""

    if (axes is None) and (ax is not None):
        axes = ax

    if (windowlabel is not None) and (fig is not None):
        fig.canvas.set_window_title(windowlabel)

    if fig is None:
        fig = _plt.gcf()

    if fig is not None and axes is None:
        axes = fig.get_axes()
        if axes == []:
            logger.error('No axes found!')

    if axes is not None:
        if toplabel is not None:
            axes.set_title(toplabel)
        if xlabel is not None:
            axes.set_xlabel(xlabel)
        if ylabel is not None:
            axes.set_ylabel(ylabel)
        if zlabel is not None:
            axes.set_zlabel(zlabel)

    if (clabel is not None) or (cb is not None):
        if (clabel is not None) and (cb is not None):
            cb.set_label(clabel)
        else:
            if clabel is None:
                logger.error('Missing colorbar label')
            else:
                logger.error('Missing colorbar instance')



# import matplotlib.pyplot as _plt

import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.widgets as _wdg
    import matplotlib as _mpl
    import numpy as _np
    import matplotlib.pyplot as _plt

from .setup_figure import setup_figure   # noqa
from .imshow import imshow
from .imshow import scaled_figsize


class Imshow_Slider(object):
    """
    .. versionchanged:: 1.2
       Name changed, colorbar options added, *p* changed to :class:`AxesImage <scisalt.matplotlib.Imshow_Slider.AxesImage>`.

    Convenience class for viewing images.
    
    Plots *image* to a to an instance of :class:`matplotlib.axis.imshow(**kwargs)`, with sliders for controlling bounds, with *\*\*kwargs* passed through to :meth:`matplotlib.axes.Axes.imshow`.

    *usecbar* determines if a colorbar will be used. Color bars can slow down the viewer significantly.
    """
    def __init__(self, image, usecbar=False, **kwargs):
        # ======================================
        # Save input info
        # ======================================
        self._image  = image
        self._kwargs = kwargs
        self.usecbar = usecbar

        # ======================================
        # Create figure
        # ======================================
        # self.fig, self.gs = setup_figure(20, 10, figsize=scaled_figsize(self.image))
        self.fig, self.gs = setup_figure(20, 10)
        self._ax_img = self.fig.add_subplot(self.gs[0:-3, :])
        self.ax_min = self.fig.add_subplot(self.gs[-2, 1:-1])
        self.ax_max = self.fig.add_subplot(self.gs[-1, 1:-1])

        self._reset(**kwargs)

    def _reset(self, **kwargs):
        # ======================================
        # Strip kwargs for vmin, vmax
        # in order to set sliders correctly
        # ======================================
        minslide = kwargs.get('vmin', self.imgmin)
        maxslide = kwargs.get('vmax', self.imgmax)

        # ======================================
        # Slider Logic
        # ======================================
        if self.imgmin > 0:
            slidermin = kwargs.pop('smin', 0)
        else:
            slidermin = kwargs.pop('smin', self.imgmin)

        if self.imgmax < 0:
            slidermax = kwargs.pop('smax', 0)
        else:
            slidermax = kwargs.pop('smax', self.imgmax)

        # ======================================
        # Imshow
        # ======================================
        # self._AxesImage = self.ax.imshow(self.image, **kwargs)
        self._AxesImage = imshow(self.image, ax=self.ax, **kwargs)
        
        shape_x, shape_y = self.image.shape
        ratio = shape_x / shape_y
        figsize = kwargs.get('figsize', _mpl.rcParams['figure.figsize'])
        figsize_ratio = figsize[0]/figsize[1]
        if ratio > figsize_ratio:
            x_lim = [0, shape_x]
            y_lim = shape_x / figsize_ratio
            y_lim = [(shape_y-y_lim)/2, (shape_y+y_lim)/2]
        else:
            x_lim = shape_y * figsize_ratio
            x_lim = [(shape_x-x_lim)/2, (shape_x+x_lim)/2]
            y_lim = [0, shape_y]

        self.ax.set_xlim(x_lim)
        self.ax.set_ylim(y_lim)

        # ======================================
        # Add minimum slider
        # ======================================
        self.minslider = _wdg.Slider(self.ax_min, 'Min', slidermin, slidermax, minslide)

        # ======================================
        # Add maximum slider
        # ======================================
        self.maxslider = _wdg.Slider(self.ax_max, 'Max', slidermin, slidermax, maxslide, slidermin=self.minslider)
        self.minslider.slidermax = self.maxslider

        self.minslider.on_changed(self._update_clim)
        self.maxslider.on_changed(self._update_clim)

        if self.usecbar:
            self.fig.colorbar(self.AxesImage, ax=self.ax, use_gridspec=True)

        self.fig.tight_layout()

    def set_cmap(self, cmap):
        """
        Sets color map to *cmap*.
        """
        self.AxesImage.set_cmap(cmap)

    @property
    def AxesImage(self):
        """
        The :class:`matplotlib.image.AxesImage` from :meth:`matplotlib.axes.Axes.imshow`.
        """
        return self._AxesImage

    @property
    def ax(self):
        """
        The :class:`matplotlib.axes.Axes` used for :meth:`matplotlib.axes.Axes.imshow`.
        """
        return self._ax_img

    # ======================================
    # Get min of image
    # ======================================
    @property
    def imgmax(self):
        """
        Highest value of input image.
        """
        return _np.max(self.image)

    # ======================================
    # Get max of image
    # ======================================
    @property
    def imgmin(self):
        """
        Lowest value of input image.
        """
        return _np.min(self.image)

    # ======================================
    # Update the clims
    # ======================================
    def _update_clim(self, val):
        cmin = self.minslider.val
        cmax = self.maxslider.val
        # print('Cmin: {}, Cmax: {}'.format(cmin, cmax))
        self.AxesImage.set_clim(cmin, cmax)

    # ======================================
    # Easily get and set slider
    # ======================================
    @property
    def clim_min(self):
        """
        Slider value for minimum
        """
        return self.minslider.val

    @clim_min.setter
    def clim_min(self, val):
        self.minslider.set_val(val)

    @property
    def clim_max(self):
        """
        Slider value for maximum
        """
        return self.maxslider.val

    @clim_max.setter
    def clim_max(self, val):
        self.maxslider.set_val(val)

    @property
    def image(self):
        """
        The image loaded.
        """
        return self._image



#!/usr/bin/env python
import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.pyplot as _plt

from .addlabel import addlabel as _addlabel
from .figure import figure as _figure


def plot_featured(*args, **kwargs):
    """
    Wrapper for matplotlib.pyplot.plot() / errorbar().

    Takes options:

    * 'error': if true, use :func:`matplotlib.pyplot.errorbar` instead of :func:`matplotlib.pyplot.plot`. *\*args* and *\*\*kwargs* passed through here.
    * 'fig': figure to use.
    * 'figlabel': figure label.
    * 'legend': legend location.
    * 'toplabel': top label of plot.
    * 'xlabel': x-label of plot.
    * 'ylabel': y-label of plot.

    """
    # Strip off options specific to plot_featured
    toplabel = kwargs.pop('toplabel', None)
    xlabel   = kwargs.pop('xlabel', None)
    ylabel   = kwargs.pop('ylabel', None)
    legend   = kwargs.pop('legend', None)
    error    = kwargs.pop('error', None)
    # save     = kwargs.pop('save', False)
    figlabel = kwargs.pop('figlabel', None)
    fig      = kwargs.pop('fig', None)

    if figlabel is not None:
        fig = _figure(figlabel)
    elif fig is None:
        try:
            fig = _plt.gcf()
        except:
            fig = _plt.fig()

    # Pass everything else to plot
    if error is None:
        _plt.plot(*args, **kwargs)
    else:
        _plt.errorbar(*args, **kwargs)

    # Format plot as desired
    _addlabel(toplabel, xlabel, ylabel, fig=fig)
    if legend is not None:
        _plt.legend(legend)

    return fig



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np

import logging
loggerlevel = 9
logger      = logging.getLogger(__name__)

__all__ = ['BDES2K', 'K2BDES']


def BDES2K(bdes, quad_length, energy):
    """
    Returns the geometric strength for a quadrupole with *bdes* and length *quad_length* for a beam with a given *energy*.
    """
    # Make sure everything is float
    bdes        = _np.float_(bdes)
    quad_length = _np.float_(quad_length)
    energy      = _np.float_(energy)

    Brho = energy/_np.float_(0.029979)
    K = bdes/(Brho*quad_length)
    logger.log(level=loggerlevel, msg='Converted BDES: {bdes}, quad length: {quad_length}, energy: {energy} to K: {K}'.format(
        bdes        = bdes        ,
        quad_length = quad_length ,
        energy      = energy      ,
        K           = K
        )
        )

    return K


def K2BDES(K, quad_length, energy):
    """
    Returns the BDES for a quadrupole with geometric strength *K* and length *quad_length* for a beam with a given *energy*.
    """
    # Make sure everything is float
    K           = _np.float_(K)
    quad_length = _np.float_(quad_length)
    energy      = _np.float_(energy)

    Brho = energy/_np.float_(0.029979)
    BDES = K*Brho*quad_length
    logger.log(level=loggerlevel, msg='Converted K: {K}, quad length: {quad_length}, energy: {energy} to BDES: {bdes}'.format(
        bdes        = BDES        ,
        quad_length = quad_length ,
        energy      = energy      ,
        K           = K
        )
        )
    return BDES



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import scipy.constants as _spc
    import numpy as _np
    import scipy as _sp

    # ============================
    # Constants
    # ============================
    e      = _spc.e               # Elementary charge
    c      = _spc.speed_of_light  # Speed of light
    e0     = _spc.epsilon_0       # Vacuum permittivity
    amu    = _spc.atomic_mass     # AMU in kg


class Ions(object):
    def __init__(self, dims, species, N_e, sig_xi, rtol=None, atol=None):
        self._dims    = dims
        self._species = species
        self._N_e     = N_e
        self._sig_xi  = sig_xi
        self._rtol = rtol
        self._atol = atol

    @property
    def rtol(self):
        return self._rtol

    @property
    def atol(self):
        return self._atol

    @property
    def species(self):
        return self._species
    
    @property
    def dims(self):
        """
        Number of dimensions
        """
        return self._dims

    @property
    def lambda_small(self):
        """
        The wavelength for small (:math:`r_0 < \\sigma_r`) oscillations.
        """
        return 2*_np.pi/_np.sqrt(self.k_small)

    # ============================
    # Omega for large q(0)
    # ============================
    def omega_big(self, q0):
        return 2*_np.pi/self.lambda_large(q0)

    @property
    def m(self):
        """
        Ion mass
        """
        return amu * self.species.mass

    @property
    def A(self):
        """
        Ion mass in units of AMU
        """
        return self.species.mass

    @property
    def N_e(self):
        """
        Number of electrons in bunch
        """
        return self._N_e

    @property
    def sig_xi(self):
        """
        Longitudinal R.M.S. width
        """
        return self._sig_xi

    def q(self, x, q0):
        """
        Numerically solved trajectory function for initial conditons :math:`q(0) = q_0` and :math:`q'(0) = 0`.
        """
        y1_0 = q0
        y0_0 = 0
        y0   = [y0_0, y1_0]

        y = _sp.integrate.odeint(self._func, y0, x, Dfun=self._gradient, rtol=self.rtol, atol=self.atol)

        return y[:, 1]

    @property
    def q_label(self):
        return 'q'



def showfig(fig, aspect="auto"):
    """
    Shows a figure with a typical orientation so that x and y axes are set up as expected.
    """

    ax = fig.gca()

    # Swap y axis if needed
    alim = list(ax.axis())
    if alim[3] < alim[2]:
        temp    = alim[2]
        alim[2] = alim[3]
        alim[3] = temp
        ax.axis(alim)

    ax.set_aspect(aspect)
    fig.show()



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import matplotlib.pyplot as _plt
    import matplotlib.gridspec as _gridspec
    import numpy as _np
    from matplotlib.backends.backend_pdf import PdfPages as _PdfPages

import logging as _logging


logger = _logging.getLogger(__name__)


def imshow_batch(images, cbar=True, show=True, pdf=None, figsize=(16, 12), rows=2, columns=2, cmap=None, **kwargs):
    """
    Plots an array of *images* to a single window of size *figsize* with *rows* and *columns*.

    * *cmap*: Specifies color map
    * *cbar*: Add color bars
    * *show*: If false, dismisses each window after it is created and optionally saved
    * *pdf*: Save to a pdf of filename *pdf*
    * *\*\*kwargs* passed to :class:`matplotlib.axis.imshow`
    """
    # ======================================
    # Set up grid
    # ======================================
    images = _np.array(images)
    gs = _gridspec.GridSpec(rows, columns)

    num_imgs = images.shape[0]
    max_ind = num_imgs-1

    # ======================================
    # Split into pages
    # ======================================
    per_page = rows*columns
    num_pages = _np.int(_np.ceil(num_imgs/per_page))
    fig_array = _np.empty(shape=num_pages, dtype=object)

    if num_pages > 1:
        logger.info('Multiple pages necessary')

    if pdf is not None:
        f = _PdfPages(pdf)

    for p in range(num_pages):
        # ======================================
        # Make figure
        # ======================================
        fig_array[p] = _plt.figure(figsize=figsize)

        # ======================================
        # Get number of rows on page
        # ======================================
        pg_max_ind = _np.min( [(p+1) * per_page - 1, max_ind] )
        num_rows = _np.int(_np.ceil((pg_max_ind+1 - p * per_page) / columns))

        for i in range(num_rows):
            # ======================================
            # Get images for column
            # ======================================
            i_min_ind = p * per_page + i * columns
            col_max_ind = _np.min([i_min_ind + columns - 1, max_ind])
            
            for j, image in enumerate(images[i_min_ind:col_max_ind+1]):
                ax = fig_array[p].add_subplot(gs[i, j])

                try:
                    if _np.issubdtype(image.dtype, _np.integer):
                        image = _np.array(image, dtype=float)
                except:
                    pass

                plot = ax.imshow(image, **kwargs)
                if cmap is not None:
                    plot.set_cmap(cmap)

                if cbar:
                    fig_array[p].colorbar(plot)

        fig_array[p].tight_layout()
        if pdf is not None:
            f.savefig(fig_array[p])

        if not show:
            _plt.close(fig_array[p])

    if pdf is not None:
        f.close()

    return fig_array
